from dataclasses import dataclass
from pathlib import Path
from typing import Generator, Iterable, Literal, Callable, Optional
import logging

import arviz as az
import numpy as np
import pandas as pd
import pymc as pm


@dataclass
class SurvivalAnalysis:
    """
    Conduct a survival analysis on an antibody dynamics run.

    Time:
        'start' and 'end' delimit the first and last time intervals used in the analysis.
        There are actually (end - start - 1) intervals in the survival analysis because
        infection risk during interval t is modelled as a function of titer at time t-1.
        (There is no risk modelled in the first time slice.)

                  `start`              `end`
                     |                   |
        titer:      [t0, t1, t2, t3, t4, t5]
                        ╲   ╲   ╲   ╲   ╲
        infection:  [i0, i1, i2, i3, i4, i5]


    Args:
        idata: InferenceData object
        start: First time slice of the analysis.
        end: Last time slice of the analysis.
        data: CombinedAllITititersData containing raw cohort data.

    Attributes:
        n_int: Number of intervals.
        intervals: end - start array containing [0, ..., n_int - 2].
        t0: Pandas month period corresponding to the first interval in the infected and
            exposure arrays (i.e. one month after `start`).
        infected: (n_ind, n_int) array of mean infection probabilities processed
            such that the maximum cumulative infection probability for any individual is
            one.
        exposure: (n_ind, n_int)
        {s,n}_titer: (n_ind, n_int)
        sn_titer: (n_ind, n_int, 2)

    """

    abd_inference: dict | az.InferenceData
    start: int
    end: int
    cohort_data: "abdpymc.TiterData"

    def __post_init__(self):
        self.from_MAP = isinstance(self.abd_inference, dict)

        self.intervals = np.arange(self.end - self.start - 1)
        self.n_int = len(self.intervals)
        self.t0 = self.cohort_data.t0 + self.start + 1

        # Infections. Transpose because inference data arrays are all (gap x ind), but
        # have implemented the survival analysis code to work on (ind x gap) arrays
        if self.from_MAP:
            infected_raw = self.abd_inference["i"].T
        else:
            self.post = az.extract(self.abd_inference)
            self.post_mean = self.post.mean(dim="sample")
            infected_raw = self.post_mean["i"].values.T

        # Remove all infection probabilities that occur after the last sample was taken
        infected_raw = make_nan_after_last_sample(
            infected_raw, last_gap=self.cohort_data.last_gap
        )

        # Extract infection probability for time gaps that are used in this survival
        # analysis.
        self.infected = infected_raw[:, self.start + 1 : self.end]

        # Using value_while_cumsum_below_threshold here means that the total infection
        # probability for each individual (row) never exceeds 1.0. I.e. at most each
        # individual is allowed a single infection.
        self.infected = np.array(
            [
                list(value_while_cumsum_below_threshold(row, threshold=1.0))
                for row in self.infected
            ]
        )

        # Exposure in the first time gap for an individual is 1.0. In subsequent gaps it
        # decreases by the cumulative sum of the infection probability from preceding
        # gaps.
        self.exposure = np.ones_like(self.infected)
        self.exposure[:, 1:] -= self.infected.cumsum(axis=1)[:, :-1]

        # Make sure that the pattern of missing data (np.nan) is the same in exposure as
        # in infected. Because exposure is calculated from the cumulative infection
        # probability that is lagged by one time gap (see above), nan's appear one time
        # step too late in the exposure array without this step.
        self.exposure[np.isnan(self.infected)] = np.nan

        # Extract antibody titer inferences. These arrays are the same size as `infected`
        # and `exposure`, but offset by one time gap such that the month `m` in the
        # titer array is aligned with month `m+1` in the exposure and infection arrays.
        if self.from_MAP:
            self.s_titer = self.abd_inference["ab_s_mu"].T[:, self.start : self.end - 1]
            self.n_titer = self.abd_inference["ab_n_mu"].T[:, self.start : self.end - 1]
        else:
            self.s_titer = self.post_mean["ab_s_mu"].values.T[
                :, self.start : self.end - 1
            ]
            self.n_titer = self.post_mean["ab_n_mu"].values.T[
                :, self.start : self.end - 1
            ]

    def model_alone(self, antigen: Literal["S", "N"]) -> pm.Model:
        """
        Bayesian cox proportional hazards model. S or N titer is used as predictor of
        infection risk in subsequent month.

        Args:
            antigen: 'S' (spike) or 'N' (nucleocapsid).
        """
        infected = convert_nan_to_zero(self.infected)
        exposure = convert_nan_to_zero(self.exposure)
        titer = {"S": self.s_titer, "N": self.n_titer}[antigen]
        coords = dict(intervals=self.intervals)
        with pm.Model(coords=coords) as model:
            a = pm.Normal("a", 0.0, 1.0)
            b = pm.Normal("b", 0.0, 1.0)
            lam0 = make_hierarchical_lam0(dims="intervals", hyper_mu=-3.0)
            lam = lam0 / (1.0 + np.exp((titer - a) * -b))
            pm.Poisson("obs", exposure * lam, observed=infected)

        return model

    @property
    def model_combined(self) -> pm.Model:
        """
        Bayesian cox proportional hazards model. S and N titer are used as predictor of
        infection risk in subsequent month.
        """
        infected = convert_nan_to_zero(self.infected)
        exposure = convert_nan_to_zero(self.exposure)
        X = np.stack((self.s_titer, self.n_titer), axis=2)
        coords = dict(intervals=self.intervals, titers=["S", "N"])
        with pm.Model(coords=coords) as model:
            a = pm.Normal("a", mu=0.0, sigma=0.5, dims="titers")
            b = pm.Normal("b", mu=0.0, sigma=0.5, dims="titers")
            lam0 = make_hierarchical_lam0(dims="intervals", hyper_mu=-3.0)
            lam = lam0 / (1.0 + np.exp(((X - a) * -b).sum(axis=2)))
            pm.Poisson("obs", exposure * lam, observed=infected)
        return model

    def model_alone_groups(
        self, antigen: Literal["S", "N"], groups: np.ndarray, group_name: str
    ) -> pm.Model:
        infected = convert_nan_to_zero(self.infected)
        exposure = convert_nan_to_zero(self.exposure)
        coords = {
            "intervals": self.intervals,
            group_name: np.sort(np.unique(groups)),
        }
        titer = {"S": self.s_titer, "N": self.n_titer}[antigen]

        with pm.Model(coords=coords) as model:
            lam0 = make_hierarchical_lam0(dims="intervals")
            a = make_hierarchical_normal("a", dims=group_name)
            b = pm.Normal("b", 0.0, 0.5)
            x = (titer - a[groups, np.newaxis]) * -b
            lam = lam0 / (1.0 + np.exp(x))
            pm.Poisson("obs", exposure * lam, observed=infected)

        return model


def value_while_cumsum_below_threshold(
    values: Iterable[float], threshold: float = 1.0
) -> Generator[float, None, None]:
    """
    Generate values while their cumulative sum is below some threshold. The (maximum) sum
    out the output is always the threshold. An example is useful:

    Example:

        values = [0.09, 0.13, 0.15, 0.03, 0.24, 0.23, 0.19, 0.03, 0.25, 0.17]
        cumsum = [0.09, 0.22, 0.37, 0.4 , 0.64, 0.87, 1.06, 1.09, 1.34, 1.51]
        output = [0.09, 0.13, 0.15, 0.03, 0.24, 0.23, 0.13, 0.00, 0.00, 0.00]
                                                        |
                                Here there is only 0.13 'left' in the cumsum (from the
                                previous cumsum of 0.87) despite the value being 0.19.

    Args:
        values:
        threshold: Maximum sum of the output.
    """
    cumsum = 0.0
    for value in values:
        if value < 0.0:
            raise ValueError(f"values must be positive: {value}")
        elif cumsum > threshold:
            yield 0.0
        else:
            # if this value takes the cumulative sum over the threshold, yield whatever
            # is 'left' in the cumulative sum to get to the threshold, otherwise just
            # yield the value
            yield threshold - cumsum if cumsum + value > threshold else value

        cumsum += value


def make_nan_after_last_sample(
    arr: np.ndarray, last_gap: dict[int, int] | pd.Series
) -> np.ndarray:
    """
    Make entries in arr that are after the last sample taken for an individual nan.

    Args:
        arr: (n_inds, n_ints)
        last_sample: Mapping of individual i -> last gap.
    """
    if np.issubdtype(arr.dtype, np.integer):
        arr = arr.astype(float)
    for ind in range(arr.shape[0]):
        if last_gap[ind] < 0:
            raise ValueError("values in last_gap must be positive")
        arr[ind, last_gap[ind] + 1 :] = np.nan
    return arr


def make_hierarchical_lam0(dims: list[str], **kwds) -> "pytensor.TensorVariable":
    """
    lam0 (lambda_0) is the baseline risk.

    Args:
        dims: Dimensions for lam0.
        **kwds passed to make_hierarchical normal.
    """
    _lam0_raw = make_hierarchical_normal(name="_lam0_raw", dims=dims, **kwds)
    return pm.Deterministic("lam0", pm.math.invlogit(_lam0_raw), dims=dims)


def make_hierarchical_normal(
    name: str,
    dims: str,
    hyper_mu: float = 0.0,
    hyper_sigma: float = 0.5,
    hyper_lam: float = 2.0,
    dims_mu: Optional[tuple[str, ...] | str] = None,
    dims_sigma: Optional[tuple[str, ...] | str] = None,
) -> "pytensor.tensor.TensorVariable":
    """
    Non-center parametrised hierarchical normal distribution. Equivalent to:

        mu = Normal(`name`_mu, mu=hyper_mu, sigma=hyper_sigma, dims=dims_mu)
        sigma = Exponential(`name`_sigma, lam=hyper_lam, dims=dims_sigma)
        Normal(name, mu=mu, sigma=sigma, dims=dims)

    Args:
        name: Variable name.
        dims: Dimensions of the model for the variable.
        hyper_{mu,sigma,lam}: Hyperpriors.
        dims_{mu,sigma}
    """
    mu = pm.Normal(f"{name}_mu", mu=hyper_mu, sigma=hyper_sigma, dims=dims_mu)
    sigma = pm.Exponential(f"{name}_sd", lam=hyper_lam, dims=dims_sigma)
    z = pm.Normal(f"_{name}_z", mu=0.0, sigma=1.0, dims=dims)
    return pm.Deterministic(name, z * sigma + mu, dims=dims)


def convert_nan_to_zero(arr: np.ndarray) -> np.ndarray:
    """Convert nan values to zero."""
    return np.where(np.isnan(arr), 0.0, arr)


def load_or_sample_model(
    path: str, model: pm.Model | Callable, *args, **kwargs
) -> az.InferenceData:
    """
    Try to load a pymc trace from a path. If the path doesn't exist then generate the
    trace by sampling from model or calling model(*args, **kwargs). The resulting
    inference data is saved to path, then returned.

    Args:
        path: Path to NetCDF file.
        model: A pymc Model or Callable that returns an arviz inference data object.
        *args, **kwargs: Passed to pymc.sample or the callable.
    """
    if not Path(path).suffix == ".nc":
        raise ValueError(f"file must end with .nc: {path}")

    if not (Path(path).parent.exists() and Path(path).parent.is_dir()):
        raise ValueError(f"directory does not exist: {path}")

    try:
        idata = az.from_netcdf(path)
        logging.info(f"loaded existing NetCDF: {path}")

    except FileNotFoundError:
        if isinstance(model, pm.Model):
            logging.info("sampling from pymc model passed")
            with model:
                idata = pm.sample(*args, **kwargs)
        else:
            logging.info("sampling by calling model")
            idata = model(*args, **kwargs)

        logging.info(f"writing new NetCDF: {path}")
        idata.to_netcdf(path)

    finally:
        return idata



from pathlib import Path
import unittest

import numpy as np
from pydantic import ValidationError

import abdpymc.simulation as sim


class TestProtection(unittest.TestCase):
    """
    Tests for abdpymc.simulation.Protection
    """

    def test_p_protection_high_titer(self):
        """
        At a very high titer, probability of protection should be very close to 1.
        """
        p = sim.Protection(a=0, b=1).p_protection(100)
        self.assertAlmostEqual(1.0, p)

    def test_p_protection_different_a(self):
        """
        Probability of protection at a given titer should be lower with a higher a (and
        positive b).
        """
        self.assertLess(
            sim.Protection(a=1, b=1).p_protection(0),
            sim.Protection(a=0, b=1).p_protection(0),
        )

    def test_p_protection_different_titer(self):
        """
        Probability of protection should be higher at a higher titer (when b is
        positive).
        """
        ag = sim.Protection(a=0, b=1)
        self.assertGreater(ag.p_protection(1), ag.p_protection(0))


class TestDynamics(unittest.TestCase):
    """
    Tests for abdpymc.simulation.Dynamics
    """

    def test_temp_wane_must_be_between_0_1(self):
        with self.assertRaisesRegex(ValidationError, "temp_wane"):
            sim.Dynamics(temp_wane=1.5)

        with self.assertRaisesRegex(ValidationError, "temp_wane"):
            sim.Dynamics(temp_wane=-0.5)

    def test_temp_rise_i_must_be_positive(self):
        with self.assertRaisesRegex(ValidationError, "temp_rise_i"):
            sim.Dynamics(temp_rise_i=-1)

    def test_perm_rise_must_be_positive(self):
        with self.assertRaisesRegex(ValidationError, "perm_rise"):
            sim.Dynamics(perm_rise=-1)

    def test_next_temp_response(self):
        ag = sim.Dynamics(temp_wane=0.95)
        temp_response = ag.next_temp_response(prev=1.0, is_infected=False)
        self.assertEqual(1.0 * 0.95, temp_response)

    def test_next_temp_response_with_infection(self):
        ag = sim.Dynamics(temp_wane=0.95, temp_rise_i=1.8)
        temp_response = ag.next_temp_response(prev=1.0, is_infected=True)
        self.assertEqual(1.0 * 0.95 + 1.8, temp_response)

    def test_cant_pass_unknown_arg(self):
        with self.assertRaisesRegex(ValidationError, "Extra inputs are not permitted"):
            sim.Dynamics(xyz=123)

    def test_perm_response_incl_vaccination(self):
        ag = sim.Dynamics(temp_wane=0.95, temp_rise_i=1.8, perm_rise=2.2)
        perm_response = ag.perm_response(
            infections=np.zeros(5), vaccinations=np.array([0, 0, 1, 0, 0])
        )
        self.assertEqual(2.2, perm_response)


class TestAntibodies(unittest.TestCase):
    """Tests for abdpymc.simulation.Antibodies"""

    def test_default_responses(self):
        """
        Should be able to make an instance without providing any values.
        """
        sim.Antibodies()

    def test_protected_n_alone(self):
        """
        Individual should be protected if N titer alone is high enough.
        """
        imm = sim.Antibodies(
            s=sim.Antibody(protection=sim.Protection(a=100, b=1)),
            n=sim.Antibody(protection=sim.Protection(a=-100, b=1)),
        )
        self.assertTrue(imm.is_protected(s_titer=0, n_titer=0))

    def test_protected_s_alone(self):
        """
        Individual should be protected if S titer alone is high enough.
        """
        imm = sim.Antibodies(
            s=sim.Antibody(protection=sim.Protection(a=-100, b=1)),
            n=sim.Antibody(protection=sim.Protection(a=100, b=1)),
        )
        self.assertTrue(imm.is_protected(s_titer=0, n_titer=0))

    def test_not_protected(self):
        """
        Individual should not be protected if both S an N titer are very low.
        """
        imm = sim.Antibodies(
            s=sim.Antibody(protection=sim.Protection(a=0, b=1)),
            n=sim.Antibody(protection=sim.Protection(a=0, b=1)),
        )
        self.assertFalse(imm.is_protected(s_titer=-100, n_titer=-100))


class TestIndividual(unittest.TestCase):
    def test_responses_default_values(self):
        """
        Should be able to instantiate without passing Responses.
        """
        sim.Individual(pcrpos=np.array([0, 0, 0, 0, 0]), vacs=np.array([0, 0, 1, 0, 0]))

    def test_vacs_pcrpos(self):
        """
        Test vacs and pcrpos attributes are passed correctly.
        """
        ind = sim.Individual(
            pcrpos=np.array([0, 0, 0, 0, 0]), vacs=np.array([0, 0, 1, 0, 0])
        )
        self.assertTrue((ind.vacs == np.array([0, 0, 1, 0, 0])).all())
        self.assertTrue((ind.pcrpos == np.array([0, 0, 0, 0, 0])).all())

    def test_cant_pass_vacs_pcrpos_diff_shape(self):
        """
        Passing vacs and pcrpos that are different shapes should raise a ValueError.
        """
        with self.assertRaisesRegex(
            ValueError, "vaccination and pcrpos are different shapes"
        ):
            sim.Individual(pcrpos=np.array([0, 0, 1]), vacs=np.array([1, 0, 0, 0]))

    def test_vacs_must_be_1d(self):
        """
        Vacs must be 1D.
        """
        with self.assertRaisesRegex(
            ValueError, "vaccination and pcrpos should be 1 dimensional"
        ):
            sim.Individual(
                pcrpos=np.array([[0, 0, 1], [0, 1, 0]]),
                vacs=np.array([[0, 0, 1], [0, 1, 0]]),
            )

    def test_pcrpos_must_be_1d(self):
        """
        pcrpos must be 1D.
        """
        with self.assertRaisesRegex(
            ValueError, "vaccination and pcrpos should be 1 dimensional"
        ):
            sim.Individual(
                pcrpos=np.array([[0, 0, 1], [0, 1, 0]]),
                vacs=np.array([[0, 0, 1], [0, 1, 0]]),
            )

    def test_infection_responses_returns_3tuple(self):
        """
        Infection responses should return a 3-namedtuple containing np.ndarrays
        """
        ind = sim.Individual(pcrpos=np.array([0, 0, 1]), vacs=np.array([0, 1, 0]))

        output = ind.infection_responses(lam0=np.array([0.1, 0.1, 0.1]))

        self.assertIsInstance(output, sim.InfectionResponses)
        self.assertEqual(3, len(output))
        self.assertIsInstance(output.s_response, np.ndarray)
        self.assertIsInstance(output.n_response, np.ndarray)
        self.assertIsInstance(output.infections, np.ndarray)

        for arr in output:
            self.assertEqual(3, len(arr))

    def test_no_responses_if_lam0_0(self):
        """
        If the baseline infection rate is always 0 then responses shouldn't change from
        initial responses.
        """
        ind = sim.Individual(
            pcrpos=np.array([0, 0, 0]),
            vacs=np.array([0, 0, 0]),
            ab=sim.Antibodies(
                s=sim.Antibody(dynamics=sim.Dynamics(init=0.123)),
                n=sim.Antibody(dynamics=sim.Dynamics(init=0.456)),
            ),
        )

        output = ind.infection_responses(lam0=np.array(np.zeros(3)))

        self.assertTrue((output.s_response == 0.123).all())
        self.assertTrue((output.n_response == 0.456).all())

    def test_pcrpos_get_recognised_as_infections(self):
        """
        A PCR+ should cause a 1 in the infections output.
        """
        ind = sim.Individual(
            pcrpos=np.array([0, 0, 1, 0, 0]), vacs=np.array([0, 0, 0, 0, 0])
        )
        output = ind.infection_responses(lam0=np.array(np.zeros(5)))
        self.assertTrue((output.infections == np.array([0, 0, 1, 0, 0])).all())

    def test_pcrpos_response_s(self):
        """
        Test S responses are correct after a PCR+.
        """
        ind = sim.Individual(
            pcrpos=np.array([0, 0, 1, 0, 0]),
            vacs=np.array([0, 0, 0, 0, 0]),
            ab=sim.Antibodies(
                s=sim.Antibody(
                    dynamics=sim.Dynamics(
                        temp_rise_i=0.3, perm_rise=0.34, temp_wane=0.94, init=-1
                    )
                )
            ),
        )

        output = ind.infection_responses(lam0=np.array(np.zeros(5)))

        # Before the PCR+
        self.assertTrue((np.array([-1, -1]) == output.s_response[:1]).all())

        # At the point of the PCR+
        self.assertAlmostEqual(-1 + 0.3 + 0.34, output.s_response[2])

        # One time step after PCR+
        self.assertAlmostEqual(-1 + 0.3 * 0.94 + 0.34, output.s_response[3])

    def test_pcrpos_response_n(self):
        """
        Test N responses are correct after a PCR+.
        """
        ind = sim.Individual(
            pcrpos=np.array([0, 0, 1, 0, 0]),
            vacs=np.array([0, 0, 0, 0, 0]),
            ab=sim.Antibodies(
                n=sim.Antibody(
                    dynamics=sim.Dynamics(
                        temp_rise_i=0.89, perm_rise=2.34, temp_wane=0.87
                    )
                )
            ),
        )

        output = ind.infection_responses(lam0=np.array(np.zeros(5)))

        # Before the PCR+
        self.assertTrue((np.array([-2, -2]) == output.n_response[:1]).all())

        # At the point of the PCR+
        self.assertAlmostEqual(-2 + 0.89 + 2.34, output.n_response[2])

        # One time step after PCR+
        self.assertAlmostEqual(-2 + 0.89 * 0.87 + 2.34, output.n_response[3])

    def test_vaccination_effect_on_s(self):
        """
        Vaccinations should increase the S response.
        """
        ind = sim.Individual(
            pcrpos=np.array([0, 0, 0, 0, 0]),
            vacs=np.array([0, 0, 1, 0, 0]),
            ab=sim.Antibodies(
                s=sim.Antibody(
                    dynamics=sim.Dynamics(
                        temp_rise_v=0.3, perm_rise=0.34, temp_wane=0.94, init=-1
                    )
                )
            ),
        )

        output = ind.infection_responses(lam0=np.array(np.zeros(5)))

        # Before the vaccination
        self.assertTrue((np.array([-1, -1]) == output.s_response[:1]).all())

        # At the point of the vaccination
        self.assertAlmostEqual(-1 + 0.3 + 0.34, output.s_response[2])

        # One time step after vaccination
        self.assertAlmostEqual(-1 + 0.3 * 0.94 + 0.34, output.s_response[3])

    def test_vaccination_and_pcrpos_effect_on_s(self):
        """
        Vaccinations should increase the S response.
        """
        ind = sim.Individual(
            pcrpos=np.array([0, 0, 0, 1, 0]),
            vacs=np.array([0, 0, 1, 0, 0]),
            ab=sim.Antibodies(
                s=sim.Antibody(
                    dynamics=sim.Dynamics(
                        temp_rise_i=0.21,
                        temp_rise_v=0.3,
                        perm_rise=0.34,
                        temp_wane=0.94,
                        init=-1,
                    )
                )
            ),
        )

        output = ind.infection_responses(lam0=np.array(np.zeros(5)))

        # Before vaccination
        self.assertTrue((np.array([-1, -1]) == output.s_response[:1]).all())

        # At the point of the vaccination
        self.assertAlmostEqual(-1 + 0.3 + 0.34, output.s_response[2])

        # One time step after vaccination, gets infected
        self.assertAlmostEqual(-1 + 0.3 * 0.94 + 0.34 + 0.21, output.s_response[3])

    def test_vaccination_and_pcrpos_effect_on_n(self):
        """
        Only PCR+ should impact N response.
        """
        ind = sim.Individual(
            pcrpos=np.array([0, 0, 0, 1, 0]),
            vacs=np.array([0, 0, 1, 0, 0]),
            ab=sim.Antibodies(
                n=sim.Antibody(
                    dynamics=sim.Dynamics(
                        temp_rise_v=0.0,
                        temp_rise_i=0.89,
                        perm_rise=2.34,
                        temp_wane=0.87,
                    )
                ),
            ),
        )

        output = ind.infection_responses(lam0=np.array(np.zeros(5)))

        # Before PCR+ all values should be initial
        self.assertTrue((np.array([-2, -2, -2]) == output.n_response[:3]).all())

        # One time step after vaccination, gets infected
        self.assertAlmostEqual(-2 + 0.89 + 2.34, output.n_response[3])


class TestCohort(unittest.TestCase):
    @classmethod
    def setUp(cls):
        cohort_data_path = Path(Path(sim.__file__).parent.parent, "data", "cohort_data")
        cls.cohort = sim.Cohort(random_seed=42, cohort_data_path=cohort_data_path)

    def test_fresh_cohort_lacks_n_titer(self):
        """
        Cohort instances shouldn't have s_titer or n_titer attributes when they are
        instantiated.
        """
        self.assertFalse(hasattr(self.cohort, "n_titer"))

    def test_fresh_cohort_lacks_s_titer(self):
        """
        Cohort instances shouldn't have s_titer or n_titer attributes when they are
        instantiated.
        """
        self.assertFalse(hasattr(self.cohort, "s_titer"))

    def test_simulate_responses_s_titer_shape(self):
        """
        Calling .simulate_responses should attach an s_titer attribute that has the same
        shape as self.vacs.
        """
        lam0 = np.repeat(0.04, self.cohort.n_gaps)
        self.cohort.simulate_responses(lam0=lam0)
        self.assertEqual(self.cohort.true.vacs.shape, self.cohort.s_titer.shape)

    def test_simulate_responses_n_titer_shape(self):
        """
        Calling .simulate_responses should attach an n_titer attribute that has the same
        shape as self.vacs.
        """
        lam0 = np.repeat(0.04, self.cohort.n_gaps)
        self.cohort.simulate_responses(lam0=lam0)
        self.assertEqual(self.cohort.true.vacs.shape, self.cohort.n_titer.shape)

    def test_simulate_responses_infections_shape(self):
        """
        Calling .simulate_responses should attach an infections attribute that has the
        same shape as self.vacs.
        """
        lam0 = np.repeat(0.04, self.cohort.n_gaps)
        self.cohort.simulate_responses(lam0=lam0)
        self.assertEqual(self.cohort.true.vacs.shape, self.cohort.infections.shape)



import unittest
from pathlib import Path

import numpy as np
import arviz as az

import abdpymc
from abdpymc.survival import SurvivalAnalysis, make_nan_after_last_sample


class TestSurvivalAnalysis(unittest.TestCase):
    """
    Tests for abdpymc.survival.SurvivalAnalysis.
    """

    @classmethod
    def setUpClass(cls):
        root = Path(abdpymc.__file__).parent.parent
        idata_path = Path(root, "data", "inference_data", "abd-20231202-256.nc")
        cls.idata = az.from_netcdf(idata_path)

        cohort_data_path = Path(root, "data", "cohort_data")
        cls.data = abdpymc.TiterData.from_disk(cohort_data_path)

    def test_make_instance(self):
        """
        Test can make an instance of SurvivalAnalysis.
        """
        SurvivalAnalysis(self.idata, start=0, end=10, cohort_data=self.data)

    def test_infected_dims(self):
        """
        sa.infected should be n_inds x n_int (1520, 10)
        """
        sa = SurvivalAnalysis(self.idata, start=0, end=10, cohort_data=self.data)
        self.assertEqual((1520, 9), sa.infected.shape)

    def test_exposure_dims(self):
        """
        sa.exposure should be n_inds x n_int (1520, 10)
        """
        sa = SurvivalAnalysis(self.idata, start=0, end=10, cohort_data=self.data)
        self.assertEqual((1520, 9), sa.exposure.shape)

    def test_dims_equal(self):
        """
        Test the dimensions of all arrays that should be equal.
        """
        sa = SurvivalAnalysis(self.idata, start=0, end=10, cohort_data=self.data)
        self.assertTrue(
            sa.exposure.shape
            == sa.infected.shape
            == sa.n_titer.shape
            == sa.s_titer.shape,
            f"shapes differ {sa.exposure.shape}, {sa.infected.shape}, "
            f"{sa.n_titer.shape}, {sa.s_titer.shape}",
        )

    def test_infected_max_sum_one(self):
        """
        The maximum sum of any row in infected should be 1.0.
        """
        sa = SurvivalAnalysis(self.idata, start=0, end=10, cohort_data=self.data)
        self.assertFalse((sa.infected.sum(axis=1) > 1.0).any())

    def test_exposure_min(self):
        """
        Minimum value of exposure should be 0.0.
        """
        sa = SurvivalAnalysis(self.idata, start=0, end=10, cohort_data=self.data)
        self.assertEqual(0.0, np.nanmin(sa.exposure))

    def test_exposure_max(self):
        """
        Maximum value of exposure should be 1.0.
        """
        sa = SurvivalAnalysis(self.idata, start=0, end=10, cohort_data=self.data)
        self.assertEqual(1.0, np.nanmax(sa.exposure))

    def test_initial_exposure_value_for_all_inds(self):
        """
        The initial value of exposure for all individuals should be 1.0.
        """
        sa = SurvivalAnalysis(self.idata, start=0, end=10, cohort_data=self.data)
        self.assertTrue((sa.exposure[:, 0] == 1.0).all())

    def test_exposure_decreasing(self):
        """
        All values in exposure should be monotonically non-increasing.
        """
        sa = SurvivalAnalysis(self.idata, start=0, end=10, cohort_data=self.data)
        diff = np.diff(sa.exposure, axis=1)
        diff_lt_zero = np.diff(sa.exposure, axis=1) <= 0.0
        diff_is_nan = np.isnan(diff)
        self.assertTrue(np.bitwise_or(diff_lt_zero, diff_is_nan).all())

    def test_infection_contains_nan(self):
        """
        Infection array should contain nan's past an individual's last sample.
        """
        sa = SurvivalAnalysis(self.idata, start=0, end=10, cohort_data=self.data)
        self.assertTrue(np.isnan(sa.infected).any())

    def test_exposure_nan_where_infection_nan(self):
        """
        The exposure array should have an identical pattern of nan values to the
        infection array.
        """
        sa = SurvivalAnalysis(self.idata, start=0, end=10, cohort_data=self.data)
        self.assertTrue(np.array_equal(np.isnan(sa.exposure), np.isnan(sa.infected)))


class TestMakeNanAfterLastSample(unittest.TestCase):
    """
    Tests for abdpymc.survival.make_nan_after_last_sample
    """

    def test_last_gap_negative(self):
        """
        Any values in last_gap that are negative should raise a ValueError.
        """
        arr = np.random.randn(10, 5)
        last_gap = dict(zip(np.arange(10), np.arange(-1, 9)))
        with self.assertRaisesRegex(ValueError, "values in last_gap must be positive"):
            make_nan_after_last_sample(arr, last_gap)

    def test_result(self):
        """
        Values after the last_gap should become nan.
        """
        arr = np.random.randn(3, 5)
        last_gap = {0: 3, 1: 0, 2: 10}

        expect = np.copy(arr)
        expect[0, 3 + 1 :] = np.nan
        expect[1, 0 + 1 :] = np.nan

        output = make_nan_after_last_sample(arr, last_gap)
        self.assertTrue(
            np.array_equal(expect, output, equal_nan=True),
            f"expect:\n{expect}\n\ndiffers to output:\n{output}",
        )



import unittest

from abdpymc.plotting import overhanging_slice


class TestOverHangingSlice(unittest.TestCase):

    """
    Tests for abdpymc.plotting.overhanging_slice
    """

    def test_case_a(self):
        self.assertEqual((5,), overhanging_slice(range(10), center=5, pad=0))

    def test_case_b(self):
        self.assertEqual((4, 5, 6), overhanging_slice(range(10), center=5, pad=1))

    def test_case_c(self):
        self.assertEqual(
            (None, None, 0, 1, 2), overhanging_slice(range(10), center=0, pad=2)
        )

    def test_case_d(self):
        self.assertEqual(
            (None, 0, 1, 2, 3), overhanging_slice(range(10), center=1, pad=2)
        )

    def test_case_e(self):
        self.assertEqual(
            (5, 6, 7, 8, 9, None, None), overhanging_slice(range(10), center=8, pad=3)
        )



import logging
from typing import Iterable, Literal, Optional

import xarray as xr
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import arviz as az

from . import abd


def overhanging_slice(values: Iterable, center: int, pad: float, fill=None) -> tuple:
    values = tuple(values)

    if center < 0:
        raise ValueError("center must be positive")

    if center > len(values):
        raise ValueError("center must be less than or equal to length of values")

    if pad < 0:
        raise ValueError("pad < 0")

    output = [fill] * (1 + 2 * pad)

    for i_output, i_values in enumerate(range(center - pad, center + pad + 1)):
        if 0 <= i_values < len(values):
            output[i_output] = values[i_values]

    return tuple(output)


def relative_to_known_ticks(
    window: int, ax: Optional[mpl.axes.Axes] = None, label: str = "PCR+"
) -> None:
    """
    Set xticks that show time of the known infection with positive and negative month
    offsets. This assumes that the known infection was in the center.

    Args:
        window: Width of the window.
        ax: Matplotlib ax.
        label: The label to put in the center.
    """
    if window % 2 != 1:
        raise ValueError("window should be odd")

    ax = plt.gca() if ax is None else ax
    labels = [""] * window

    labels[window // 2] = label

    for i, tick in enumerate(range(window // 2 + 1, window), start=1):
        labels[tick] = f"+{i}"

    for i, tick in enumerate(reversed(range(window // 2)), start=1):
        labels[tick] = f"−{i}"

    ax.set_xticks(np.arange(window), labels)


def infp_known_centered(p_infection: np.ndarray, known: np.ndarray, pad: float):
    """
    Construct an array where p_infection are aligned relative to all known infections.

    Args:
        p_infection: (n_gap, n_ind) array containing infection probabilities.
        known: (n_gap, n_ind) array of known infections. Must be same shape as
            p_infection.
        pad: Value to fill arrays if pad extends outside of p_infection.
    """
    if p_infection.shape != known.shape:
        raise ValueError(
            f"p_infection shape {p_infection.shape} != known shape {known.shape}"
        )

    return np.array(
        [
            overhanging_slice(p_infection[:, ind], center=gap, pad=pad, fill=np.nan)
            for gap, ind in np.argwhere(known)
        ]
    )


def plot_p_infection_relative_to_known_infection(
    p_infection: np.ndarray,
    known_infection: np.ndarray,
    pad: int,
    points: bool = False,
    lines: bool = False,
    mean_line: bool = False,
    cum_mean_line: bool = False,
    ax: Optional[mpl.axes.Axes] = None,
    xlabel: str = "Months since PCR+",
    xtick_center_label: str = "PCR+",
) -> None:
    """
    Plot infection probabilities relative to known infections. All known infections are
    centered in the plot and infection probabilities relative the known infections are
    shown `pad` months before and after.

    Args:
        p_infection: (n_gap, n_ind) array containing infection probabilities.
        known_infection: (n_gap, n_ind) array of known infections. Must be same shape as
            p_infection.
        pad: Value to fill arrays if pad extends outside of p_infection.
        points: Add points on lines.
        lines: Plot all lines.
        mean_line: Show the mean of all infection probabilities.
        cum_mean_line: Show the cumulative mean infection probability.
        ax: Matplotlib axes.
        xlabel: X-label.
        xtick_center_label: Label for the central xtick.
    """
    ax = plt.gca() if ax is None else ax

    infp_overlaid = infp_known_centered(
        p_infection=p_infection, known=known_infection, pad=pad
    )

    n_known, window = infp_overlaid.shape

    # each known infection gets its own level of jitter
    jitter = np.tile(np.random.uniform(-0.25, 0.25, n_known), (window, 1)).T

    x = np.arange(window)

    x_tile = np.tile(x, (n_known, 1))

    kwds = dict(clip_on=False)

    if points:
        ax.scatter(x_tile + jitter, infp_overlaid, alpha=0.1, lw=0, zorder=12, **kwds)

    if lines:
        ax.plot(
            x_tile.T + jitter.T,
            infp_overlaid.T,
            alpha=0.05,
            zorder=10,
            c="grey",
            **kwds,
        )

    if mean_line:
        ax.plot(
            x,
            np.nanmean(infp_overlaid, axis=0),
            linewidth=2,
            marker="o",
            markersize=5,
            c="black",
            zorder=15,
            label="Mean",
            **kwds,
        )

    if cum_mean_line:
        cumsum = np.nanmean(infp_overlaid, axis=0).cumsum()
        logging.info(
            "cummulative posterior mean infection probabilities:\n"
            f"{np.round(cumsum, decimals=3)}"
        )
        ax.plot(
            x,
            cumsum,
            linewidth=2,
            marker="o",
            markersize=5,
            c="black",
            linestyle=":",
            zorder=15,
            label="Cumulative mean",
            **kwds,
        )

    ax.axvline(window // 2, c="black", ls="--")

    relative_to_known_ticks(window, ax=ax, label=xtick_center_label)

    ax.set(ylim=(0, 1), ylabel="P(Infection)", xlabel=xlabel)


def plot_protection_timevariable(idata, **kwds):
    """
    Plot protection for all intervals in a time variable model.
    """
    for interval in idata.posterior.coords["intervals"].values:
        plot_protection(idata, interval=interval, **kwds)


def plot_protection(
    idata: az.InferenceData | xr.Dataset,
    n_samples=250,
    xmin=-1,
    xmax=6,
    show_mean=True,
    show_hdi: bool = False,
    antigen: Optional[Literal["S", "N"]] = None,
    interval: Optional[int] = None,
    sample_kwds: Optional[dict] = None,
    hdi_kwds: Optional[dict] = None,
    mean_kwds: Optional[dict] = None,
    ax: Optional[mpl.axes.Axes] = None,
) -> None:
    """
    Plot posterior protection curve.

    Arguments:
        idata: InferenceData object or Dataset containing the posterior samples.
        n_samples: The number of posterior samples to plot. Default is 250.
        xmin: The minimum x-axis value. Default is -1.
        xmax: The maximum x-axis value. Default is 6.
        show_mean: Whether to show the mean curve. Default is True.
        show_hdi: Whether to show the highest density interval (HDI) curve. Default is False.
        antigen: For combined models, the antigen to be plotted.
        interval: For time variable models, which interval should be plotted?
        sample_kwds: Additional keyword arguments for customizing the sample curve plot.
            Passed to plt.plot.
        hdi_kwds: Additional keyword arguments for customizing the HDI curve plot.
            Passed to plt.fill_between.
        mean_kwds: Additional keyword arguments for customizing the mean curve plot.
            Passed to plt.plot.
    """
    sample_kwds = {} if sample_kwds is None else sample_kwds
    hdi_kwds = {} if hdi_kwds is None else hdi_kwds
    mean_kwds = {} if mean_kwds is None else mean_kwds
    ax = ax or plt.gca()

    post = az.extract(idata) if isinstance(idata, az.InferenceData) else idata

    x = np.linspace(xmin, xmax)

    a = post["a"]
    b = post["b"]

    if antigen is not None:
        a = a.sel(titers=antigen)
        b = b.sel(titers=antigen)

    if interval is not None:
        a = a.sel(intervals=interval)

        if "intervals" in b.dims:
            b = b.sel(intervals=interval)

    if a.dims != ("sample",):
        raise ValueError(
            f"expected a.dims to be ('sample',) but is {a.dims}. Did you forget to "
            "specify an antigen / interval for a combined / time variable model?"
        )

    if b.dims != ("sample",):
        raise ValueError(
            f"expected b.dims to be ('sample',) but is {b.dims}. Did you forget to "
            "specify an antigen for a combined model?"
        )

    y = 1.0 - abd.logistic(x=x[:, np.newaxis], a=a.values, b=b.values, d=1.0)

    if n_samples:
        if n_samples > y.shape[1]:
            raise ValueError("asking to show more samples than there are")
        samples = np.linspace(0, y.shape[1] - 1, num=n_samples, dtype=int)
        ax.plot(
            x,
            y[:, samples],
            c=sample_kwds.pop("c", "black"),
            alpha=sample_kwds.pop("alpha", 0.05),
            **sample_kwds,
        )

    if show_hdi:
        hdi = az.hdi(y.T)
        ax.fill_between(x, hdi[:, 0], hdi[:, 1], **hdi_kwds)

    if show_mean:
        ax.plot(x, y.mean(axis=1), **mean_kwds)



#!/usr/bin/env python3

from typing import Iterable, Union, Optional, Callable, Literal
import time
import functools
import logging
import string
import warnings

import xarray as xr
import arviz as az
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl

import abdpymc as abd

Number = Union[float, int]


def grouper(iterable, n):
    """
    Inspired by:

    >>> def grouper(iterable: Iterable, n: int, fillvalue: Any = None):
    >>>    '''Group items in iterable in size n batches'''
    >>>    args = [iter(iterable)] * n
    >>>    return itertools.zip_longest(*args, fillvalue=fillvalue)

    The above grouper pads batches with fillvalue, e.g.:

    >>> list(grouper("David Pattinson", 6))
    >>> [('D', 'a', 'v', 'i', 'd', ' '),
    >>> ('P', 'a', 't', 't', 'i', 'n'),
    >>> ('s', 'o', 'n', None, None, None)]

    Wheras this version does not pad the final group with a fillvalue.
    """
    values = tuple(iterable)
    start = 0
    while start < len(values):
        yield values[start : start + n]
        start += n


def _hi_lo_scaler_scalar(p: Number, lo: Number, hi: Number):
    """Scale a proportion p between lo and hi."""
    return p * (hi - lo) + lo


def hi_lo_scaler(p: Union[Number, Iterable[Number]], lo: Number, hi: Number):
    """
    Plotting helper function.

    Scale a proportion p between the values lo and hi.
    """
    try:
        return [_hi_lo_scaler_scalar(_p, lo, hi) for _p in p]
    except TypeError:
        return _hi_lo_scaler_scalar(p, lo, hi)


def load_ititers(path: str, data: abd.TiterData) -> pd.DataFrame:
    """Load DataFrame containing inflection titers"""
    df = pd.read_csv(path, index_col=0)
    df["elapsed_months"] = [data.date_to_gap(date) for date in df["collection_date"]]
    return df


def gap_jitter(i: int, j: int) -> np.array:
    """
    Plotting helper funciton.

    Add some horizontal jitter to antibody responses over time.

    Args:
        i: Number of rows.
        j: Number of columns.

    Returns:
        (i, j) array. Each column has the same amount of jitter added. Values of each
            row grow by 1 each row. E.g.:

            >>> gap_jitter(4, 6)
            array([[0.46, 0.81, 0.25, 0.48, 0.28, 0.31],
                   [1.46, 1.81, 1.25, 1.48, 1.28, 1.31],
                   [2.46, 2.81, 2.25, 2.48, 2.28, 2.31],
                   [3.46, 3.81, 3.25, 3.48, 3.28, 3.31]])
    """
    return (
        np.random.uniform(-0.5, 0.5, j).reshape(-1, 1)
        + np.tile(np.arange(0.5, i + 0.5), j).reshape(j, i)
    ).T


def plot_events(
    events_binary: np.ndarray,
    ymin: Union[float, int],
    ymax: Union[float, int],
    **kwds,
):
    """
    Plot PCR+ or vaccination events as bars.

    Args:
        events_binary: (n_gaps,) Array containing 1s in the gaps where the events
            occurred.
        ymin: Bottom of the bar.
        ymax: Top of the bar.
        **kwds: Passed to plt.vlines.
    """
    events, *_ = np.nonzero(events_binary)
    plt.bar(
        events,
        height=ymax - ymin,
        width=1,
        bottom=ymin,
        align="edge",
        zorder=0,
        **kwds,
    )


def plot_variant_background(
    start: pd.Period,
    t0: pd.Period,
    width: int,
    ax=None,
    label=None,
    fontsize=6,
    ymin=0,
    ymax=1,
    bleed=0.75,
    **kwds,
):
    """
    Shaded rectangle showing when a particular variant circulated.

    Args:
        start: The first month this variant became predominant.
        t0: The first month on the ax.
        width: Width of the rectangle
        ax:
        label: Text label for the patch.
        fontsize:
        ymin:
        ymax:
        bleed: Amount the patch extends underneath the xaxis.
        **kwds: Passed to mpl.patches.Rectangle.
    """
    ax = plt.gca() if ax is None else ax
    x = (start - t0).n

    ax.add_artist(
        mpl.patches.Rectangle(
            xy=(x, ymin - bleed),
            width=width,
            height=ymax - ymin + bleed,
            clip_on=False,
            zorder=-1,
            **kwds,
        )
    )

    if label:
        ax.text(
            x + width / 2,
            ymax + 0.02,
            label,
            ha="center",
            va="bottom",
            zorder=20,
            fontsize=fontsize,
        )


plot_omicron_background = functools.partial(
    plot_variant_background,
    start=pd.Period("2022-01"),
    width=11,
    facecolor="#e0deed",
    label="Omicron",
    bleed=0,
)


plot_delta_background = functools.partial(
    plot_variant_background,
    start=pd.Period("2021-07"),
    width=6,
    facecolor="#fee7cd",
    label="Delta",
    bleed=0,
)


plot_predelta_background = functools.partial(
    plot_variant_background,
    start=pd.Period("2020-05"),
    width=14,
    facecolor="#daf1ed",
    label="pre-Delta",
    bleed=0,
)

plot_preomicron_background = functools.partial(
    plot_variant_background,
    start=pd.Period("2020-05"),
    width=20,
    facecolor="#daf1ed",
    label="pre-Omicron",
    bleed=0,
)


def plot_individual(
    ind_i,
    data: abd.TiterData,
    post: xr.Dataset,
    df_ititers: pd.DataFrame,
    ymin: Union[float, int] = -4,
    ymax: Union[float, int] = 8,
    post_skip: Optional[int] = None,
    splits: Optional[tuple[int]] = None,
    post_mean: Optional[xr.DataArray] = None,
    show_s_post: bool = True,
    show_n_post: bool = True,
    show_s_data: bool = True,
    show_n_data: bool = True,
    show_inf_prob: bool = True,
    show_cuminf_prob: bool = True,
    label_fontsize: float = 6,
    show_only_1_40_without_ititers: bool = True,
):
    """
    Plot posterior distribution of antibody responses for individual i in post.

    Args:
        ind_i: The index of the individual to plot in post.
        data: Data used to run ab dynamics model.
        post: Posterior distribution. chain and draw should be combined, e.g.:
            idata.posterior.stack({"sample": ("chain", "draw")})
        df_ititers: DataFrame containing inflection titers to plot.
        ymin: Minimum y value.
        ymax: Maximum y value.
        post_skip: Skip this many lines when plotting raw samples from the posterior
            distribution. If None, then calculate a value such that 50 posterior samples
            are plotted.
        splits: Plot cummulative infection probabilities within the time chunks delimited
            by splits.
        post_mean: Precomputed posterior mean. Can save time to precompute over multiple
            individuals if plot_individual is being called a lot.
        show_s_post: Show posterior infection probabilities and antibody responses for S.
        show_n_post: Show posterior infection probabilities and antibody responses for N.
        show_s_data: Show inflection titer estimates for S.
        show_n_data: Show inflection titer estimates for N.
        show_inf_prob: Stair plot showing monthly infection probability.
        show_cuminf_prob: Stair plot showing cumulative monthly infection probability.
        label_fontsize: Fontsize for labelling variant periods, individual record number
            and index.
        show_only_1_40_without_ititers: Only show 1:40 data where there isn't inflection
            titer data too.
    """
    ax = plt.gca()

    scale = functools.partial(hi_lo_scaler, lo=ymin, hi=ymax)

    record_id = data.ind_i_to_record_id[ind_i]
    df_ind_i = data.df.query("individual_i == @ind_i").set_index("elapsed_months")

    last_gap = data.last_gap[ind_i]

    post_skip = len(post.sample) // 250 if post_skip is None else post_skip

    if show_s_post or show_n_post:
        ags = {(True, False): "s", (False, True): "n", (True, True): "sn"}[
            (show_s_post, show_n_post)
        ]
        for ag in ags:
            dic = getattr(data.plot_conf, ag)
            color = dic["c"]
            it_name = dic["it"]  # 'ab_s_mu' or 'ab_n_mu'

            ititer = getattr(post, it_name)
            response = ititer[:last_gap, ind_i]

            # Plot mean response
            response_mean = (
                getattr(post_mean, it_name)[:last_gap, ind_i]
                if post_mean is not None
                else response.mean(dim="sample")
            )
            gap_mids = np.arange(0.5, last_gap + 0.5)
            ax.plot(gap_mids, response_mean, c="white", lw=2, zorder=12)
            ax.plot(gap_mids, response_mean, c=color, lw=2, ls="--", zorder=14)

            # Posterior samples
            y = response[:, ::post_skip]
            x = gap_jitter(*y.shape)
            ax.plot(x, y, c=color, lw=0.5, alpha=0.1)

    # Stairs showing monthly infection probability
    if show_inf_prob or show_cuminf_prob:
        inf_p = (
            post.sel(gap=slice(None, last_gap), ind=ind_i).mean(dim="sample").i
            if post_mean is None
            else post_mean.sel(gap=slice(None, last_gap), ind=ind_i).i
        )

    if show_inf_prob:
        ax.stairs(
            scale(inf_p),
            edges=np.arange(last_gap + 2),
            color="black",
            baseline=None,
            clip_on=False,
            zorder=10,
        )

    if show_cuminf_prob:
        cum_p = abd.compute_chunked_cum_p(inf_p, splits=splits)
        plt.stairs(
            scale(cum_p),
            edges=np.arange(0, len(cum_p) + 1),
            baseline=None,
            color="grey",
            clip_on=False,
            zorder=9,
            lw=0.5,
        )

    # PCR+ and vaccinations
    kwds = dict(ymin=ymin, ymax=ymax)
    plot_events(data.pcrpos[ind_i], color="#f2aaa8", **kwds)
    plot_events(data.vacs[ind_i], color="#afd0aa", **kwds)

    # Background patches
    kwds = dict(t0=data.t0, ymin=ymin, ymax=ymax, alpha=0.25, fontsize=label_fontsize)
    plot_preomicron_background(**kwds)
    plot_omicron_background(**kwds)

    # ELISA titers
    if show_s_data or show_n_data:
        ind_i_samples = df_ind_i["sample"].unique()
        ind_i_ititer_samples = df_ititers.index.intersection(ind_i_samples)

        kwds = dict(
            ax=ax,
            df=df_ititers.loc[ind_i_ititer_samples],
            distinguish_single_vs_multiple=False,
            hdi=True,
        )

        if show_s_data:
            plot_titers(antigen="10222020-S", color=data.plot_conf.s["c"], **kwds)

        if show_n_data:
            plot_titers(antigen="40588-V08B", color=data.plot_conf.n["c"], **kwds)

    # Individual index and record ID
    ax.text(
        0,
        1.01,
        f"i:{ind_i} r:{record_id}",
        transform=ax.transAxes,
        fontsize=label_fontsize,
    )

    # Finish ax
    ax.set_ylim(ymin, ymax)
    ax.set_xlim(0, data.n_gaps)

    finish_timeline_ax(ax, data)

    return ax


def plot_titers(
    df: pd.DataFrame,
    antigen: Literal["10222020-S", "40588-V08B"],
    ax: mpl.axes.Axes,
    color: str,
    distinguish_single_vs_multiple: bool = False,
    hdi: bool = True,
):
    """
    Plot titers inferred independently of the antibody dynamics framework.

    Args:
        df: DataFrame containing the titers.
        antigen: The catalog number of the antigen to plot.
        ax: Matplotlib ax.
        color: Color for the data.
        distinguish_single_multiple: Use different markers for titers that were inferred
            using a single dilution vs. those inferred using multiple dilutions.
        hdi: Show the HDI.
    """
    kwds = dict(c=color, lw=0.5, zorder=15)

    df = df.query(f"antigen == '{antigen}'")

    if distinguish_single_vs_multiple:
        df_single = df.query("dilutions == '40'")
        ax.scatter(
            df_single["elapsed_months"] + 0.5,
            df_single["titer"],
            marker="x",
            s=5,
            **kwds,
        )

        df_multiple = df.query("dilutions != '40'")
        ax.scatter(
            df_multiple["elapsed_months"] + 0.5,
            df_multiple["titer"],
            marker="o",
            s=20,
            ec="white",
            **kwds,
        )
    else:
        ax.scatter(
            df["elapsed_months"] + 0.5,
            df["titer"],
            marker="o",
            s=20,
            ec="white",
            **kwds,
        )

    if hdi:
        ax.errorbar(
            x=df["elapsed_months"] + 0.5,
            y=df["titer"],
            yerr=np.stack([df["titer"] - df["hdi_lo"], df["hdi_hi"] - df["titer"]]),
            fmt="none",
            elinewidth=0.5,
            c=color,
            alpha=0.5,
            capsize=1.0,
            capthick=0.5,
        )


def finish_timeline_ax(ax: mpl.axes.Axes, data: abd.TiterData, yaxis=True) -> None:
    """
    Adjust ticks and labels for a plot which has cohort time on the x-axis.

    Args:
        ax: Matplotlib ax.
        data: Cohort data object.
        yaxis: Whether to setup the y-axis to display titers.
    """
    ax.grid(zorder=0)
    ax.xaxis.set_minor_locator(mpl.ticker.IndexLocator(base=3, offset=2))
    ax.xaxis.set_major_locator(mpl.ticker.IndexLocator(base=12, offset=2))
    ax.tick_params(which="minor", length=4)
    ax.tick_params(which="major", length=8, axis="x")
    if yaxis:
        ax.set_yticks(data.plot_conf.yticks, data.plot_conf.yticklabels)
    ax.set_xticks(data.plot_conf.xticks, data.plot_conf.xticklabels)
    ax.minorticks_off()


def plot_multiple_individuals(
    nrows: int,
    ncols: int,
    ax_width: int = 4,
    individuals: Optional[tuple[int, ...]] = None,
    subplots_kwds: Optional[dict] = None,
    lettered_labels: Optional[bool] = False,
    **kwds,
):
    """
    Plot a grid of antibody timeseries for multiple individuals.

    Args:
        nrows: Number of rows.
        ncols: Number of cols
        ax_width: Width of an individual plot.
        individuals: Indexes of individuals to plot.
        subplots_kwds: Dict of keyword arguments to pass to plt.subplots.
        lettered_labels: Add letter labels to the top left of each ax.
        **kwds passed to plot_individual.
    """
    width = ax_width * ncols
    height = ax_width / 1.618 * nrows

    fig, _ = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        sharex=True,
        sharey=True,
        figsize=(width, height),
        **subplots_kwds,
    )

    individuals = range(nrows * ncols) if individuals is None else individuals

    for i, ind_i in enumerate(individuals):

        ax = fig.axes[i]

        plt.sca(ax)
        plot_individual(ind_i=ind_i, **kwds)

        if lettered_labels:
            ax.text(
                0.02,
                0.98,
                string.ascii_uppercase[i],
                weight="bold",
                transform=ax.transAxes,
                va="top",
                fontsize=16,
            )


def plot_multiple_individuals_batches(
    individuals: Iterable[int], fname: str, nrows: int = 10, ncols: int = 25, **kwds
):
    """
    Plot multiple individual timelines in batches of up to 250.

    Args:
        individuals: Indexes of individuals to plot.
        fname: Should contain two '{}' to for the first and last individual in the group.
        nrows: Number of rows of axes.
        ncols: Number of columns of axes.
        **kwds: Passed to plot_multiple_individuals.
    """
    batch_size = nrows * ncols

    for batch in grouper(individuals, batch_size):
        plot_multiple_individuals(nrows=nrows, ncols=ncols, individuals=batch, **kwds)
        plt.savefig(fname.format(batch[0], batch[-1]), bbox_inches="tight", dpi=300)
        plt.close()


def log_time(func: Callable) -> Callable:
    """Timing decorator"""

    def wrapped(*args, **kwargs):
        t0 = time.time()
        logging.log(logging.INFO, msg=f"Calling {func.__name__}", end=" ")
        result = func(*args, **kwargs)
        t1 = time.time()
        logging.log(logging.INFO, msg=f"total = {t1 - t0:2.1f} s")
        return result

    return wrapped


@log_time
def stack_posterior(idata: az.InferenceData) -> xr.Dataset:
    """Combine chain and draw dimensions into single sample dimension."""
    return idata.posterior.stack({"sample": ("chain", "draw")})


@log_time
def compute_posterior_mean(post: xr.Dataset) -> xr.Dataset:
    """Compute mean accross samples."""
    return post.mean(dim="sample")


def main():
    import argparse

    parser = argparse.ArgumentParser(
        "abdpymc-plot-timelines",
        description="Default behaviour is to plot all individuals in batches of 250. "
        "Use --individuals to plot only specific individuals or --pcrpos_only to plot "
        "only individuals that have a PCR+ test result.",
    )
    parser.add_argument("--idata", help="Path of .nc file", required=True)
    parser.add_argument(
        "--fname_pref",
        help="Prefix of PDF filenames to save. Files will have this suffix appended: "
        "'-{}-{}.pdf'. The index of the first and last individuals occupy the curly "
        "braces.",
    )
    parser.add_argument(
        "--cohort_data",
        help="Path to directory for generating CombinedAllITitersData object "
        "(default=cohort_data).",
        default="cohort_data",
        required=True,
    )
    parser.add_argument(
        "--no_cumm_p",
        action="store_true",
        help="Don't plot cumulative infection probabilities. --split_{delta,omicron} "
        "have no effect if this is passed.",
    )
    parser.add_argument(
        "--split_delta",
        action="store_true",
        help="Plot cumulative infection probabilities in separate chunks. Use this "
        "flag to split delta from pre-delta.",
    )
    parser.add_argument(
        "--split_omicron",
        action="store_true",
        help="Like --split_delta but for splitting omicron from delta.",
    )
    parser.add_argument(
        "--individuals",
        help="Plot timelines for these specific individuals.",
        nargs="+",
        type=int,
    )
    parser.add_argument(
        "--nrows",
        help="Number of rows when passing specific individuals (default=19).",
        type=int,
        default=19,
    )
    parser.add_argument(
        "--ncols",
        help="Number of columns when passing specific individuals (default=16).",
        type=int,
        default=16,
    )
    parser.add_argument(
        "--fname",
        help="Name(s) of file(s) to save when passing specific individuals. Can "
        "specify multiple names to plot, say PNG and PDF.",
        nargs="+",
    )
    parser.add_argument(
        "--pcrpos_only",
        help="Only plot individuals who have a PCR+ result. Ignored if --individuals "
        "is passed.",
        action="store_true",
    )
    parser.add_argument(
        "--lettered_labels",
        action="store_true",
        help="If using --individuals, add alphabetical labels to each ax.",
    )
    parser.add_argument(
        "--dont_show_independent_titers",
        help="Don't plot independent titer estimates stored in <cohort_data>/ititers.csv",
        action="store_true",
    )
    args = parser.parse_args()

    idata = az.from_netcdf(args.idata)

    data = abd.TiterData.from_disk(args.cohort_data)

    splits = (
        data.calculate_splits(delta=args.split_delta, omicron=args.split_omicron)
        if not args.no_cumm_p
        else None
    )

    post = stack_posterior(idata)

    post_mean = compute_posterior_mean(post)

    try:
        df_ititers = load_ititers(path=f"{args.cohort_data}/ititers.csv", data=data)
    except FileNotFoundError:
        df_ititers = None

    kwds = dict(
        data=data,
        post=post,
        post_mean=post_mean,
        splits=splits,
        df_ititers=df_ititers,
        subplots_kwds=dict(gridspec_kw=dict(wspace=0.05, hspace=0.15)),
        show_s_data=not args.dont_show_independent_titers,
        show_n_data=not args.dont_show_independent_titers,
    )

    if args.individuals:
        if len(args.individuals) != args.nrows * args.ncols:
            warnings.warn(
                f"""
                Mismatch between number of individuals ({len(args.individuals)}) and
                number of rows ({args.nrows}) and columns ({args.ncols}). Use --nrows and
                --ncols.
                """
            )

        plot_multiple_individuals(
            individuals=args.individuals,
            nrows=args.nrows,
            ncols=args.ncols,
            lettered_labels=args.lettered_labels,
            **kwds,
        )
        for path in args.fname:
            plt.savefig(path, dpi=300, bbox_inches="tight")
        plt.close()

    elif args.pcrpos_only:
        plot_multiple_individuals_batches(
            individuals=np.flatnonzero(data.pcrpos.any(axis=1)),
            fname=f"{args.fname_pref}-{{}}-{{}}.pdf",
            **kwds,
        )

    else:
        plot_multiple_individuals_batches(
            individuals=range(data.n_inds),
            fname=f"{args.fname_pref}-{{}}-{{}}.pdf",
            nrows=args.nrows,
            ncols=args.ncols,
            **kwds,
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse

import numpy as np
import arviz as az


def main():
    parser = argparse.ArgumentParser("abdpymc-subsample-idata")
    parser.add_argument(
        "--idata", required=True, help="NetCDF file of InferenceData object."
    )
    parser.add_argument("--output", required=True, help="Name of NetCDF path to write.")
    parser.add_argument(
        "--n", required=True, type=int, help="Number of samples per chain to subsample."
    )
    args = parser.parse_args()

    idata = az.from_netcdf(args.idata)

    draws = idata.sample_stats.dims["draw"]

    if args.n > draws:
        raise ValueError(
            f"asking for {args.n} samples but there are only {draws} in total"
        )

    samples = np.round(np.linspace(start=0, stop=draws - 1, num=args.n)).astype(int)

    az.to_netcdf(idata.sel(draw=samples), args.output)


if __name__ == "__main__":
    main()



from importlib.metadata import version, PackageNotFoundError

from .abd import (
    BLUEGREY,
    TiterData,
    compute_chunked_cum_p,
    DARKORANGE,
    incorporate_pcrpos,
    invlogistic,
    logistic,
    mask_future_infection,
    mask_multiple_infections_2_chunks,
    mask_multiple_infections_3_chunks,
    mask_multiple_infections,
    mask_three_gaps,
    model,
    scalar_variables,
    TwoTimeChunks,
)
from . import abd
from . import plotting
from . import simulation
from . import timelines
from . import survival
from .survival import load_or_sample_model, SurvivalAnalysis


try:
    __version__ = version("abdpymc")
except PackageNotFoundError:
    pass

__all__ = [
    "abd",
    "BLUEGREY",
    "TiterData",
    "compute_chunked_cum_p",
    "DARKORANGE",
    "incorporate_pcrpos",
    "invlogistic",
    "load_or_sample_model",
    "logistic",
    "mask_future_infection",
    "mask_multiple_infections_2_chunks",
    "mask_multiple_infections_3_chunks",
    "mask_multiple_infections",
    "mask_three_gaps",
    "model",
    "plotting",
    "scalar_variables",
    "simulation",
    "survival",
    "SurvivalAnalysis",
    "timelines",
    "TwoTimeChunks",
]



#!/usr/bin/env python3

import math
import os
from abc import ABC
from typing import Generator, Iterable, Optional, Union
import itertools

import arviz as az
import numpy as np
import pandas as pd
import pymc as pm
import pytensor as pt
import pytensor.tensor as at

DARKORANGE = "#E14D2A"
BLUEGREY = "#3E6D9C"
GAP_IND = "gap", "ind"
IND_GAP = "ind", "gap"


class AntigenTiterData:
    def __init__(self, ag: str, df: pd.DataFrame) -> None:
        """
        Titer data for a single antigen.

        Args:
            ag: One letter name of antigen e.g. S/N.
            df: DataFrame.
        """
        self.ag = ag
        self.df = df

        # Indexes to map inflection titers to the correct gaps and individuals
        self.idx_gap = df["elapsed_months"].values
        self.idx_ind = df["individual_i"].values

        # pymc distributions can't take np.int64 as shape parameters, hence int here
        self.n_gaps = int(max(self.idx_gap) + 1)
        self.n_inds = int(max(self.idx_ind) + 1)

    def __repr__(self) -> str:
        return f"AntigenTiterData(ag={self.ag}, df={self.df})"


class TiterData:
    def __init__(
        self,
        t0: pd.Period,
        df: pd.DataFrame,
        vacs: np.ndarray,
        pcrpos: np.ndarray,
        ageenroll: Optional[dict] = None,
    ) -> None:
        """
        Manage OD, PCR+ and vaccination data for an antibody dynamics analysis.

        Args:
            t0: The fist month in the analysis. All gap indexes are relative to this
                first month.
            df: See to_disk and from_disk.
            vacs: Similar to df, but for the vaccination data. See to_disk and from_disk.
            pcrpos: Similar to df, but for the PCR+ data. See to_disk and from_disk.
            ageenroll: Dict containing enrollment age.

        Attributes:
            df: All the data. sample_i enumerates unique samples.
            df_s: Only the S1S2 data.
            df_n: Only the N data.

        In df_s and df_n the sample_{s,n}_i column enumerates unique samples that have
        been measured for each antigen.
        """

        self.df = df

        # Map individual i to a record_id
        self.ind_i_to_record_id = dict(
            self.df[["individual_i", "record_id"]].drop_duplicates().values
        )

        self.s = AntigenTiterData(
            ag="s",
            df=(
                self.df.query("measurement == '10222020-S'")
                .copy()
                .assign(sample_s_i=lambda df: unique_ids(df["sample"]))
            ),
        )

        self.n = AntigenTiterData(
            ag="n",
            df=(
                self.df.query("measurement == '40588-V08B'")
                .copy()
                .assign(sample_n_i=lambda df: unique_ids(df["sample"]))
            ),
        )

        self.t0 = t0
        self.n_gaps = max(self.df["elapsed_months"]) + 1
        self.n_inds = max(self.df["individual_i"]) + 1

        self.ind_ids = list(
            tuple(v)
            for _, v in self.df[["individual_i", "record_id"]]
            .drop_duplicates()
            .iterrows()
        )
        self.record_ids = np.array([ids[1] for ids in self.ind_ids])

        # The following 2 arrays are (n_inds, n_gaps) that contain 1's if an event
        # (vaccination or PCR+) happened to an individual in a month, or 0 otherwise.
        self.vacs = vacs
        self.pcrpos = pcrpos

        self.plot_conf = PlotConfig(self)

        # gap index of the last sample for each individual
        self.last_gap = (
            self.df.groupby("individual_i")
            .aggregate({"elapsed_months": "max"})
            .squeeze()
        )

        self.coords = dict(ind=np.arange(self.n_inds), gap=np.arange(self.n_gaps))

        if ageenroll is not None:
            self.ageenroll = np.array(
                [ageenroll[record_id] for record_id in self.record_ids]
            )

    def __repr__(self) -> str:
        return f"TiterData(t0={self.t0}, n_inds={self.n_inds})"

    @property
    def periods(self) -> pd.Series:
        """
        Month periods that correspond to each gap.
        """
        return pd.Series([self.t0 + i for i in range(self.n_gaps)])

    def date_to_gap(self, period: Union[str, pd.Period]) -> int:
        """
        Which gap did a particular date occur in?
        """
        return (pd.Period(period, freq="M") - self.t0).n

    def to_disk(self, directory: str) -> None:
        """
        Write a copy of self.df, self.vacs and self.pcrpos to a directory.

        Args:
            directory
        """
        if not os.path.exists(directory):
            os.mkdir(directory)

        fmt = "%1.0f"

        def path(x):
            return f"{directory}/{x}"

        self.df.to_csv(path("df.csv"))
        np.savetxt(path("vacs.txt"), self.vacs, fmt=fmt)
        np.savetxt(path("pcrpos.txt"), self.pcrpos, fmt=fmt)

        with open(path("t0.txt"), "w") as fobj:
            fobj.write(str(self.t0))

    @classmethod
    def from_disk(cls, directory: str) -> "TiterData":
        """
        Generate an instance of CombinedAllITitersData by reading data from disk,
        rather than making calls to the mfsera package.

        Args:
            directory: Directory containing df.csv, vacs.txt, pcrpos.txt and t0.txt.
                See to_disk.
        """

        def path(x):
            return f"{directory}/{x}"

        df = pd.read_csv(path("df.csv"), index_col=0)
        vacs = np.loadtxt(path("vacs.txt"))
        pcrpos = np.loadtxt(path("pcrpos.txt"))

        try:
            ageenroll = pd.read_csv(
                path("individuals.csv"), header=None, index_col=0
            ).squeeze()
        except FileNotFoundError:
            ageenroll = None

        if vacs.shape != pcrpos.shape:
            raise ValueError("vacs and pcrpos are different shapes")

        with open(path("t0.txt"), "r") as fobj:
            t0 = pd.Period(fobj.readline())

        return cls(t0=t0, df=df, vacs=vacs, pcrpos=pcrpos, ageenroll=ageenroll)

    def calculate_splits(self, delta: bool, omicron: bool) -> tuple[int]:
        """
        Given the time of the first gap, t0, return how many gaps exist between when
        delta and/or omicron started to circulate.

        Args:
            delta: Include the number of gaps to when delta started to circulate.
            omicron: Include the number of gaps to when omicron started to circulate.
        """
        splits = []

        if delta:
            splits.append((pd.Period("2021-07") - self.t0).n)

        if omicron:
            splits.append((pd.Period("2022-01") - self.t0).n)

        return tuple(splits)


def make_decay_design(n_gaps):
    """
    Make a decay design matrix for calculating the waning exponent given an infection in each row.
    Each row contains the waning parameter exponent given an infection in a different time gap.
    Values in the lower left triangle of the array should be ignored.

    Example:

        >>> make_design(5).eval()
        array([[0, 1, 2, 3, 4],
            [0, 0, 1, 2, 3],
            [0, 0, 0, 1, 2],
            [0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0]])
    """
    return at.maximum(0, at.arange(n_gaps) - at.arange(n_gaps)[:, None])


def _temp_response_scalar_rho(exposure, n_inds, temp, rho):
    """
    Compute a temporary response using a scalar rho.

    Args:
        exposure: (n_gaps, n_inds) array of exposures.
        n_inds: Number of individuals in the dataset.
        temp: The magnitude of the temporary response.
        rho: Gap to gap waning proportion.
    """
    n_gaps = int(exposure.shape.eval()[0])
    design = make_decay_design(n_gaps)
    offset = at.tril(at.ones_like(design), -1)

    # This is an (n_gap, n_gap) array that contains the temp response given an exposure in each time
    # gap. The first row captures an exposure in the first gap, second in the second etc...
    responses_each_gap = (rho**design - offset) * temp

    return (responses_each_gap[:, :, None] * exposure[:, None, :]).sum(axis=0)


def _temp_response_vector_rho(exposure, n_inds, temp, rho):
    """
    Compute a temporary response using a vector rho. I.e. in this version each individual can have
    their own value of rho.
    """
    n_gaps = int(exposure.shape.eval()[0])
    design = make_decay_design(n_gaps)
    offset = at.tril(at.ones_like(design), -1)

    return ((rho ** design[..., None] - offset[..., None]) * exposure[:, None, :]).sum(
        axis=0
    )


def _temp_response_scan(exposure, n_inds, temp, rho):
    """
    Compute a temporary response using pytensor.tensor.scan.

    Args:
        exposure: (n_gaps, n_inds) matrix of infections, vaccinations or combination.
        n_inds: Number of individuals in the dataset.
        temp: Variable determining the size of the temporary response.
        rho: Gap to gap waning proportion.
    """
    response, _ = pt.scan(
        lambda e, prev, temp, rho: prev * rho + e * temp,
        sequences=[exposure],
        outputs_info=at.zeros(n_inds),
        non_sequences=[temp, rho],
    )
    return response


def perm_response(exposure, perm):
    """
    Compute a permanent response.

    The value is perm after the first exposure.

    Args:
        exposure: (n_gaps, n_inds) matrix of infections, vaccinations or combination.
        perm: The size of the permanent response.
    """
    return at.switch(at.cumsum(exposure, axis=0) > 0.0, perm, 0.0)


def model_n_response(data: AntigenTiterData, i: at.TensorLike):
    """
    Model an N response.

    Features:
        - Only impacted by infection events.
        - Single temporary response parameter.
        - Single waning parameter.
        - All individuals wane.

    Args:
        data: Must have n_inds, idx_gap and idx_ind attributes.
        i: (n_gap, n_ind) array containing 0/1 indicating infection events.
    """
    assert data.ag == "n"

    def var(x):
        return f"ab_{data.ag}_{x}"

    # generate a value of perm for all gaps after the first infection in each individual
    perm = pm.Gamma(var("perm"), mu=2.0, sigma=0.5)
    perm_gap_ind = perm_response(exposure=i, perm=perm)

    # response that decays over time
    temp = pm.Gamma(var("temp"), mu=1.0, sigma=0.5)
    rho = pm.Beta(var("rho"), alpha=10.0, beta=1.0)

    temp_gap_ind = _temp_response_scalar_rho(
        exposure=i, n_inds=data.n_inds, temp=temp, rho=rho
    )

    init = pm.Normal(var("init"), -2, 1)
    mu = pm.Deterministic(var("mu"), perm_gap_ind + temp_gap_ind + init, dims=GAP_IND)

    return mu[data.idx_gap, data.idx_ind]


def model_s_response(data: AntigenTiterData, i: at.TensorLike, v: at.TensorLike):
    """
    Model an S response.

    Features:
        - Impacted by infection and vaccination events.
        - Separate temporary response parameters for vaccinations and infections.
        - Single waning parameter.
        - Individuals either wane (ab_n_waner=0) or do not (ab_n_waner=1)

    Args:
        data: Must have n_inds, idx_gap and idx_ind attributes.
        i: (n_gap, n_ind) array containing 0/1 indicating infection events.
        v: (n_gap, n_ind) array containing 0/1 indicating vaccinations.
    """
    assert data.ag == "s"

    def var(x):
        return f"ab_{data.ag}_{x}"

    # Permanent response after initial exposure (vaccination or infection)
    perm = pm.Gamma(var("perm"), mu=2.0, sigma=0.5)
    perm_gap_ind = perm_response(exposure=i + v, perm=perm)

    # response that decays over time for vaccinations and infections
    rho = pm.Beta(var("rho"), alpha=10.0, beta=1.0)
    p_waner = pm.Beta(var("p_waner"), alpha=1.0, beta=1.0)  # baseline p(waner)
    waner = pm.Bernoulli(var("waner"), p=p_waner, dims="ind")
    rho_per_ind = rho * waner + 1 - waner  # waner=1 -> rho=rho, waner=0 -> rho=1

    # Infection response
    tempinf = pm.Gamma(var("tempinf"), mu=1.0, sigma=0.5)
    tempinf_gap_ind = _temp_response_vector_rho(
        exposure=i, n_inds=data.n_inds, temp=tempinf, rho=rho_per_ind
    )

    # Vaccination response
    tempvac = pm.Gamma(var("tempvac"), mu=1.0, sigma=0.5)
    tempvac_gap_ind = _temp_response_vector_rho(
        exposure=v, n_inds=data.n_inds, temp=tempvac, rho=rho_per_ind
    )

    init = pm.Normal(var("init"), -2, 1)
    mu = pm.Deterministic(
        var("mu"), perm_gap_ind + tempinf_gap_ind + tempvac_gap_ind + init, dims=GAP_IND
    )

    return mu[data.idx_gap, data.idx_ind]


def model(
    data: TiterData,
    splits: Optional[Union[tuple[int], tuple[int, int]]] = None,
    ignore_pcrpos: int = False,
):
    """
    Set up an antibody dynamics model.

    Args:
        data: Combined constant gap data.
        splits: A len 1 or len 2 tuple containing gap indexes for infection chunking.
        ignore_pcrpos: Should PCR+ data be ignored (e.g. for running a positive control
            test).
    """
    check_splits(splits=splits, data=data)

    # Vaccinations
    v = at.as_tensor(data.vacs.T)

    # PCR positives
    pcrpos = (
        at.zeros_like(data.pcrpos.T) if ignore_pcrpos else at.as_tensor(data.pcrpos.T)
    )

    time_chunks = make_time_chunks(pcrpos=pcrpos, splits=splits)

    with pm.Model(coords=data.coords) as model:
        # Base infection probability. Loosely expect approx 1 infection per person.
        p = pm.Beta("p", alpha=1, beta=data.n_gaps - 1)

        # Exposure matrix
        i_raw = pm.Bernoulli("i_raw", p, dims=GAP_IND)

        # Constrain infections if need be w.r.t. how many infections are allowed per
        # time chunk, whether PCR+ infections take precedence over inferred infections
        # and prevent infections from occurring within 3 months of another infection.
        i = time_chunks.constrain_infections(i_raw)

        # Infer infections and ab parameters
        ititers_n = model_n_response(data=data.n, i=i)
        ititers_s = model_s_response(data=data.s, i=i, v=v)

        # Sigmoid likelihood
        model_sigmoids(data=data.n, a=ititers_n)
        model_sigmoids(data=data.s, a=ititers_s)

    return model


def model_sigmoids(data: AntigenTiterData, a: at.TensorLike) -> None:
    """
    Model sigmoid curves of each sample in data. Must be called from within a pymc model
    context.

    Args:
        data: AllITiters data. a: Inflection titers inferred from infection and antibody
        dynamics component of
            the model.
    """

    def var(x):
        return f"it_{data.ag}_{x}"

    pm.Normal(
        var("lik"),
        mu=logistic(
            x=data.df["log_dilution"].values,
            a=a,
            b=pm.Normal(var("b"), -1, 0.5),
            d=pm.Normal(var("d"), 2, 0.5),
        ),
        sigma=pm.Exponential(var("sigma"), 1),
        observed=data.df["od"].values,
    )


def unique_ids(values: Iterable) -> list[int]:
    """Give a unique integer starting from 0 to each value in values.

    Args:
        values
    """
    unique_values = set(values)
    mapping = dict(zip(unique_values, range(len(unique_values))))
    return [mapping[value] for value in values]


class PlotConfig:
    def __init__(self, data: TiterData):
        """
        Commonly used attributes in plotting individual timeseries.
        """
        self.s = dict(it="ab_s_mu", c=DARKORANGE)
        self.n = dict(it="ab_n_mu", c=BLUEGREY)
        self.yticks = np.arange(0, 6, 2)
        self.yticklabels = [
            str(int(t)) for t in numeric_to_titer(self.yticks, start=40, fold=4)
        ]
        self.xticks, self.xticklabels = year_tick_labels(data.t0, data.n_gaps)


@np.vectorize
def numeric_to_titer(n, start, fold=4):
    """
    Convert a dilution to its concentration relative to the starting
    mixture based on its position in the dilution series.

    Notes:
        Imagine starting with a dilution of 1:40. If you conducted a
        four-fold dilution series the concentration would be:

        Concentration (parts) = 1:40, 1:160, 1:640, 1:2560, ...
        Position (n)          =    0,     1,     2,      3, ...

    Args:
        n (int): Position in dilution series (0-indexed).
        start (int): Initial dilution.
        fold (int): Fold difference of each subsequent dilution in series.
    """
    t = start * float(fold) ** n
    integer, fractional = math.modf(t)
    if fractional < 1e-6:
        return int(integer)
    else:
        return t


def months_from(t0: pd.Period, n: int) -> Generator[pd.Period, None, None]:
    """
    Generate n month periods starting from t0.

    Args:
        t0: First period.
        n: Number of periods to generate.
    """
    curr = pd.Period(t0)
    for _ in range(n):
        yield curr
        curr += 1


def year_tick_labels(t0: pd.Period, n_months: int) -> tuple[np.ndarray, np.ndarray]:
    """
    Generate tick locations and labels to mark years starting from t0 for
    n_months.
    """
    ticks, labels = zip(
        *(
            (i, period.year)
            for i, period in enumerate(months_from(t0, n_months))
            if period.month == 1
        )
    )
    return np.array(ticks), np.array(labels)


def invlogistic(x, a, b, d):
    return a - np.log(d / x - 1) / b


def logistic(x, a, b, d):
    return d / (1 + np.exp(-b * (x - a)))


def mask_three_gaps(arr: at.TensorLike) -> at.TensorLike:
    """
    Prevent an infection from occurring within three months of another infection.

    Args:
        arr: (n_gaps, n_inds) tensor containing 1 if an individual was infected in a
            particular time gap.
    """
    _, n_inds = arr.shape.eval()
    taps = [-3, -2, -1]
    initial = at.zeros((len(taps), n_inds), dtype="int8")

    masked, _ = pt.scan(
        mask_future_infection,
        sequences=dict(input=at.cast(arr, "int8")),
        outputs_info=dict(taps=taps, initial=initial),
    )

    return masked


def mask_future_infection(i0, im3, im2, im1):
    """
    Prevent infection if infections have occurred in any of the previous
    three months.

    All arguments are ints (0/1) that indicate whether an infection occurred in
    a certain timeslot.

    This function is passed to pytensor.scan. scan generates the next value based
    on the current value (i0) and 'taps' into other slots of the sequence. If

        -2 -1  0 +1 +2 +3
    i = [a, b, c, d, e, f]

    Args:
        i0: Current infections
        im3: Infection 3 months ago
        im2: Infection 2 months ago
        im1: Infection 1 month ago
    """
    return at.switch(im3 | im2 | im1, 0, i0)


def check_splits(
    splits: Union[tuple[int, ...], None], data: Optional[TiterData] = None
):
    """
    Check splits are valid.
    """
    if splits is not None:
        if any(split < 0 for split in splits):
            raise ValueError("split indexes must be positive")
        if sorted(splits) != list(splits):
            raise ValueError("splits must be in ascending order")
        if data is not None and splits and splits[-1] > data.n_gaps:
            raise ValueError(
                f"largest split must be less than n_gaps - 1, ({splits[-1]})"
            )
        if len(splits) != len(set(splits)):
            raise ValueError("splits not unique")
        if any(not isinstance(split, int) for split in splits):
            raise ValueError("splits must be ints")


class OneTimeChunk:
    def __init__(self, pcrpos: at.TensorLike) -> None:
        """
        A class that handles a single time chunk.

        In a single time chunk multiple infections are allowed, and PCR+ data do not
        take any special precedence over probabilistically inferred infections.

        Implementation note:
            This class does not need to inherit from the MultipleTimeChunks ABC as its
            constrain_infections method does not use mask_multiple_infections and
            incorporate_pcrpos.
        """
        self.pcrpos = pcrpos

    def constrain_infections(self, i_raw: at.TensorLike) -> at.TensorLike:
        # For a single time chunk, incorporating PCR positive data is as simple as
        # adding these two arrays.
        i_pcrpos = i_raw + self.pcrpos

        # It could be that 1s in i_raw and self.pcrpos collide
        # Here, ensure max value in i_pcrpos is 1
        i_pcrpos_max1 = at.where(i_pcrpos > 0.0, 1.0, 0.0)

        return pm.Deterministic("i", mask_three_gaps(i_pcrpos_max1), dims=GAP_IND)


class MultipleTimeChunks(ABC):
    """
    An abstract base class that defines how probabilistically inferred infections and
    PCR+ data are constrained.
    """

    def constrain_infections(self, i_raw: at.TensorLike) -> at.TensorLike:
        # Prevent multiple exposures in a single chunk
        i_raw = self.mask_multiple_infections(i_raw)

        # Include PCR+ in infection matrix
        i_raw = self.incorporate_pcrpos(i_raw)

        # Finally, prevent infections occurring within three months of each other,
        # regardless of which time chunk they occur in
        return pm.Deterministic("i", mask_three_gaps(i_raw), dims=GAP_IND)


class TwoTimeChunks(MultipleTimeChunks):
    def __init__(self, split: int, pcrpos: at.TensorLike) -> None:
        """
        A class for handling time broken into 2 different chunks. The two chunks are
        specified by passing an int.
        """
        if not isinstance(split, int):
            raise ValueError("1 split is required to specify 2 time chunks")
        else:
            self.split = split

        n_gaps, n_inds = pcrpos.shape.eval()
        self.n_gaps = int(n_gaps)
        self.n_inds = int(n_inds)

        self.pcrpos_0 = pcrpos[:split]
        self.pcrpos_1 = pcrpos[split:]

    def mask_multiple_infections(self, arr):
        return mask_multiple_infections_2_chunks(arr, split=self.split)

    def incorporate_pcrpos(self, i_raw):
        return at.concatenate(
            (
                incorporate_pcrpos(i_raw[: self.split], self.pcrpos_0),
                incorporate_pcrpos(i_raw[self.split :], self.pcrpos_1),
            )
        )


class ThreeTimeChunks(MultipleTimeChunks):
    def __init__(self, splits: tuple[int], pcrpos: at.TensorLike):
        """
        A class for handling time broken into 3 different chunks. The three chunks are
        specified by passing a tuple of 2 ints.
        """
        check_splits(splits)

        if len(splits) != 2:
            raise ValueError("2 splits are required to specify 3 time chunks")
        else:
            self.split0, self.split1 = splits

        self.pcrpos_0 = pcrpos[: self.split0]
        self.pcrpos_1 = pcrpos[self.split0 : self.split1]
        self.pcrpos_2 = pcrpos[self.split1 :]

    def mask_multiple_infections(self, arr):
        return mask_multiple_infections_3_chunks(
            arr, split0=self.split0, split1=self.split1
        )

    def incorporate_pcrpos(self, i_raw):
        return at.concatenate(
            (
                incorporate_pcrpos(i_raw[: self.split0], self.pcrpos_0),
                incorporate_pcrpos(i_raw[self.split0 : self.split1], self.pcrpos_1),
                incorporate_pcrpos(i_raw[self.split1 :], self.pcrpos_2),
            )
        )


def incorporate_pcrpos(i_raw: at.TensorLike, pcrpos: at.TensorLike):
    """
    Incorporate PCR positive data into raw infection.

    Raw infections are a matrix of 1s and 0s. Each row is a time gap, each column is an
    individual:

                i_raw = [
                    [0, 1, 0, 0],
                    [0, 0, 0, 0],
                    [1, 1, 0, 0],
                    [0, 0, 0, 1],
                    [0, 0, 0, 0],
                    [0, 1, 1, 0],
                    [0, 0, 0, 0],
                ]

    PCR+ is encoded as a matrix of the same shape where a 1 indicates a PCR+ for a
    particular individual at a particular time:

                pcrpos = [
                    [0, 0, 0, 0],
                    [0, 0, 0, 0],
                    [1, 0, 0, 0],
                    [0, 0, 0, 0],
                    [0, 0, 0, 0],
                    [0, 0, 1, 0],
                    [0, 0, 0, 0],
                ]

    This function returns another matrix of the same shape. When an individual has a PCR+
    the column from the PCR+ is taken. When they don't have a PCR+ the column from i_raw
    is taken. This effectively removes any other infections for an individual when it is
    known they had a PCR+ in a particular time gap.

    Args:
        i_raw: (gap, ind)
        pcrpos: (gap, ind)
    """
    return at.where(pcrpos.any(axis=0), pcrpos, i_raw)


def mask_multiple_infections_3_chunks(arr, split0, split1):
    """
    A version of mask_multiple_infections_2_chunks, but with 2 splits / 3 chunks.

    Args:
        arr: The array containing infections.
        split0: The index where the second chunk begins.
        split1: The index where the third chunk begins.
    """
    return at.concatenate(
        (
            mask_multiple_infections(arr[:split0]),
            mask_multiple_infections(arr[split0:split1]),
            mask_multiple_infections(arr[split1:]),
        )
    )


def mask_multiple_infections(arr: at.TensorLike):
    """
    Prevent successive infections occurring.

    The following matrix encodes when individuals where infected. Each column is a
    different individual, each row is a different time gap.

    [
        [1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1],
        [1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 0, 0],
        [1, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 0, 1],
        [0, 1, 0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0]
    ]

    This function turns 1s -> 0s if they occur after another 1 in a column:

    [
        [1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0],
        [0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 1],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    ]

    Args:
        arr: (n gap, n individual) matrix encoding when an individual was infected.
    """
    return at.where(arr.cumsum(axis=0) > 1, 0, arr)


def mask_multiple_infections_2_chunks(arr: at.TensorLike, split: int):
    """
    A single infection is allowed within each chunk of the array.
    The first infection in a chunk takes precedence.

    E.g. if this array were split into two chunks:

       [
           [1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1],
           [1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 0, 0],
           [1, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 0, 1],
           [0, 1, 0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0],  Chunk 1
           #------------------------------------------------------
           [1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0],  Chunk 2
           [0, 1, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1],
           [0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0],
           [1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0]
       ]

    This should be the result:

    [
        [1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1],
        [0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  Chunk 1
        #------------------------------------------------------
        [1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0],  Chunk 2
        [0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 1],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    ]


    Args:
           arr: The array containing infections.
           split: The index where the second chunk begins. E.g. in the example above the
            index where chunk 2 starts is 4.
    """
    return at.concatenate(
        (mask_multiple_infections(arr[:split]), mask_multiple_infections(arr[split:]))
    )


def make_time_chunks(
    pcrpos: at.TensorLike,
    splits: Optional[Union[tuple[int], tuple[int, int]]] = None,
) -> OneTimeChunk | TwoTimeChunks | ThreeTimeChunks:
    """
    Helper function to make a TimeChunks class based on the number of splits.
    """
    if splits is None or len(splits) == 0:
        return OneTimeChunk(pcrpos=pcrpos)

    elif len(splits) == 1:
        return TwoTimeChunks(split=splits[0], pcrpos=pcrpos)

    elif len(splits) == 2:
        return ThreeTimeChunks(splits=splits, pcrpos=pcrpos)

    else:
        raise NotImplementedError("only implemented 1-3 time chunks (0-2 splits)")


def main():
    import argparse

    parser = argparse.ArgumentParser("abdpymc-infer")
    parser.add_argument(
        "--tune", help="Number of tuning steps.", type=int, required=True
    )
    parser.add_argument("--draws", help="Number of draws.", type=int, required=True)
    parser.add_argument("--cores", help="Number of cores", type=int)
    parser.add_argument(
        "--ititers_data",
        help="Path to directory for generating TiterData object.",
        default="cohort_data",
    )
    parser.add_argument(
        "--split_delta",
        help="Split time chunk between delta and pre-delta",
        action="store_true",
    )
    parser.add_argument(
        "--split_omicron",
        help="Split time chunk between omicron and delta",
        action="store_true",
    )
    parser.add_argument("--ignore_pcrpos", help="Ignore PCR+ data", action="store_true")
    parser.add_argument("--netcdf", help="Path of netCDF file to save.")
    args = parser.parse_args()

    data = TiterData.from_disk(args.ititers_data)

    splits = (
        None
        if (not args.split_delta) and (not args.split_omicron)
        else data.calculate_splits(delta=args.split_delta, omicron=args.split_omicron)
    )

    with model(data, splits=splits, ignore_pcrpos=args.ignore_pcrpos):
        idata = pm.sample(tune=args.tune, draws=args.draws, cores=args.cores)

    az.to_netcdf(idata, args.netcdf)


def scalar_variables(dataset: "xarray.Dataset") -> list[str]:
    """List variables of an xarray dataset that only have a 'sample' dimension."""
    return [var for var in dataset.data_vars if dataset[var].dims == ("sample",)]


def compute_chunked_cum_p(
    p: np.array, splits: Optional[tuple[int, ...]] = None
) -> np.array:
    """
    Compute cumulative probabilities from an array of probabilities. Time is split into
    chunks defined by splits. Compute cumulative probabilities for each chunk.

    Args:
        p: 1D array containing probabilities.
        splits: Defines the borders of different chunks of time in p.
    """
    if p.ndim != 1:
        raise ValueError("p is not 1D")

    if splits is None:
        cum = np.cumsum(p)
        return np.where(cum > 1.0, 1.0, cum)
    else:
        return np.hstack(
            [
                compute_chunked_cum_p(p[a:b])
                for a, b in itertools.pairwise((0, *splits, None))
            ]
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import logging
import unittest
from pathlib import Path

import numpy as np
import pytensor.tensor as at
import pymc as pm

import abdpymc as abd


class TestMaskFutureInfections(unittest.TestCase):
    """Tests for abd.mask_future_infection"""

    def test_no_prior_infections(self):
        """Infection should be allowed if there are no prior infections"""
        self.assertEqual(abd.mask_future_infection(1, 0, 0, 0).eval(), 1)

    def test_no_prior_infection_no_current_infection(self):
        """Check infections don't appear from nowhere."""
        self.assertEqual(abd.mask_future_infection(0, 0, 0, 0).eval(), 0)

    def test_current_value_is_returned(self):
        """
        Whatever is in the "i0" slot should be returned, if there are no
        prior infections.
        """
        self.assertEqual(abd.mask_future_infection(10, 0, 0, 0).eval(), 10)

    def test_infection_in_slot_m3_masks(self):
        """An infection 3 slots before should mask a current infection."""
        self.assertEqual(abd.mask_future_infection(1, 1, 0, 0).eval(), 0)

    def test_infection_in_slot_m2_masks(self):
        """An infection 2 slots before should mask a current infection."""
        self.assertEqual(abd.mask_future_infection(1, 0, 1, 0).eval(), 0)

    def test_infection_in_slot_m1_masks(self):
        """An infection 1 slots before should mask a current infection."""
        self.assertEqual(abd.mask_future_infection(1, 0, 0, 1).eval(), 0)

    def test_infection_in_slot_m3_doesnt_create_infection(self):
        """
        An infection 3 slots before current should not produce an
        infection.
        """
        self.assertEqual(abd.mask_future_infection(0, 1, 0, 0).eval(), 0)

    def test_infection_in_slot_m2_doesnt_create_infection(self):
        """
        An infection 2 slots before current should not produce an
        infection.
        """
        self.assertEqual(abd.mask_future_infection(0, 0, 1, 0).eval(), 0)

    def test_infection_in_slot_m1_doesnt_create_infection(self):
        """
        An infection 1 slots before current should not produce an
        infection.
        """
        self.assertEqual(abd.mask_future_infection(0, 0, 0, 1).eval(), 0)


class TestAllITitersData(unittest.TestCase):
    """
    Tests for abd.AllITitersData
    """

    @classmethod
    def setUp(cls):
        """
        Build an instance of CombinedAllITitersData for testing. This object has 's' and
        'n' attributes which are instances of AllITitersData.
        """
        root = Path(abd.__file__).parent.parent
        directory = root.joinpath(Path("data", "cohort_data"))
        cls.cat = abd.TiterData.from_disk(directory)

    def test_idx_gap_shape(self):
        """
        Index to extract values of a from (n_gaps, n_inds) matrix should be same length
        as the data.
        """
        self.assertEqual(len(self.cat.n.idx_gap), len(self.cat.n.df))

    def test_idx_ind_shape(self):
        """
        Index to extract values of a from (n_gaps, n_inds) matrix should be same length
        as the data.
        """
        self.assertEqual(len(self.cat.n.idx_ind), len(self.cat.n.df))


class TestMaskInfectionsWithin3Months(unittest.TestCase):
    """
    Tests for functions that prevent infections occurring immediately after one another.
    When an infection occurs, an infection is not allowed to occur in the following
    three time chunks.
    """

    def test_mask_infections_3_months_shape(self):
        """
        The output shape should match the input shape.
        """
        input = at.as_tensor(
            [
                [1, 1, 0],
                [1, 0, 0],
                [1, 0, 0],
                [1, 0, 0],
                [1, 0, 0],
            ]
        )
        output = abd.mask_three_gaps(input).eval()
        self.assertEqual((5, 3), output.shape)

    def test_mask_infections_within_3_months(self):
        """
        An infection should not be possible 3 months after another infection.
        """
        input = at.as_tensor_variable(
            [
                [1, 1, 0],
                [1, 0, 0],
                [1, 0, 0],
                [1, 0, 0],
                [1, 0, 0],
            ]
        )
        output = abd.mask_three_gaps(input).eval()
        expect = at.as_tensor_variable(
            [
                [1, 1, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [1, 0, 0],
            ]
        )
        nrows, ncols = input.shape.eval()
        for i in range(nrows):
            for j in range(ncols):
                with self.subTest(i=i, j=j):
                    self.assertEqual(expect[i, j].eval(), output[i, j])


class TestMaskMultipleInfections(unittest.TestCase):
    """
    Tests for functionality around constraining indiviudals to only be allowed a single
    infection per variant.

    mask_multiple_infections takes a tensor like:

    [
        [0, 1, 1, 0, 0, 1],
        [1, 1, 0, 0, 1, 0],
        [0, 1, 1, 0, 0, 1]
    ]

    And turns any 1 that is preceded by another 1 in a column to 0.
    """

    @classmethod
    def setUp(cls):
        """
        An array for testing.
        """
        np.random.seed(42)
        cls.n_gap = 29
        cls.n_ind = 50
        cls.arr = at.as_tensor_variable(
            np.random.randint(0, 2, size=(cls.n_gap, cls.n_ind))
        )

    def test_returns_same_shape(self):
        """
        Should return the same shape array that is passed to it.
        """
        out = abd.mask_multiple_infections(self.arr)
        self.assertEqual((29, 50), tuple(out.shape.eval()))

    def test_col_max_1(self):
        """
        The maximum sum of any column in the returned array should be 1.
        """
        out = abd.mask_multiple_infections(self.arr)
        colsums = out.sum(axis=0).eval()
        self.assertEqual(1, max(colsums))

    def test_returns_0_or_1(self):
        """
        Only 0 or 1 are allowed values in the returned array.
        """
        out = abd.mask_multiple_infections(self.arr)
        self.assertEqual({0, 1}, set(out.eval().ravel()))

    def test_exact_value(self):
        """
        Test the correct value is returned.
        """
        input = np.array([[0, 1, 1, 0, 0, 1], [1, 1, 0, 0, 1, 0], [0, 1, 1, 0, 0, 1]])
        expect = np.array([[0, 1, 1, 0, 0, 1], [1, 0, 0, 0, 1, 0], [0, 0, 0, 0, 0, 0]])
        out = abd.mask_multiple_infections(input).eval()
        self.assertTrue(np.all(np.equal(expect, out)))


class TestMaskMultipleInfectionsChunks(unittest.TestCase):
    """
    Tests to check functionality where masking multiple infections is split into
    different chunks. A single infection is allowed within each chunk of the array.
    The first infection in a chunk takes precedence.

    E.g. if this array were split into two chunks:

    [
        [1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1],
        [1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 0, 0],
        [1, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 0, 1],
        [0, 1, 0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0],  Chunk 1
        #------------------------------------------------------
        [1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0],  Chunk 2
        [0, 1, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1],
        [0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0],
        [1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0]
    ]

    Then this should be the product:

    [
        [1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1],
        [0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],  Chunk 1
        #------------------------------------------------------
        [1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0],  Chunk 2
        [0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 1],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    ]
    """

    @classmethod
    def setUp(cls):
        """
        An array for testing.
        """
        np.random.seed(42)
        cls.n_gap = 29
        cls.n_ind = 50
        cls.arr = at.as_tensor_variable(
            np.random.randint(0, 2, size=(cls.n_gap, cls.n_ind))
        )

    def test_returns_same_shape(self):
        """
        Should return the same shape array that is passed to it.
        """
        out = abd.mask_multiple_infections_2_chunks(self.arr, split=18)
        self.assertEqual((29, 50), tuple(out.shape.eval()))

    def test_col_max_2(self):
        """
        The maximum sum of any column in the returned array should be 2.
        """
        out = abd.mask_multiple_infections_2_chunks(self.arr, split=18)
        colsums = out.sum(axis=0).eval()
        self.assertEqual(2, max(colsums))

    def test_returns_0_or_1(self):
        """
        Only 0 or 1 are allowed values in the returned array.
        """
        out = abd.mask_multiple_infections_2_chunks(self.arr, split=18)
        self.assertEqual({0, 1}, set(out.eval().ravel()))

    def test_exact_value_single_split(self):
        """
        Test the correct value is returned for the single split case.
        """
        input = [
            [1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 0, 0],
            [1, 1, 0, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 0, 0],
            [1, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 0, 0],
            [0, 1, 0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0],
            # ---------------------------------------------
            [1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 0, 1, 0],
            [0, 1, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0],
            [1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0],
        ]
        expect = [
            [1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 0, 0, 0],
            [0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            # ------------------------------------y--------
            [1, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 0, 1, 0],
            [0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ]

        out = abd.mask_multiple_infections_2_chunks(
            at.as_tensor_variable(input), split=4
        ).eval()
        print(out)
        self.assertTrue(np.all(np.equal(expect, out)))

    def test_exact_value_double_split(self):
        """
        Test the correct value is returned for the double split case.
        """
        input = at.as_tensor_variable(
            [
                [1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 1],
                [0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1],
                [0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 1, 1, 0, 1, 0],
                [0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 1],
                # ---------------------------------------------
                [0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 0],
                [0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0],
                [0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 1, 1, 1, 1],
                [1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1],
                [1, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 0, 1, 1, 1],
                [1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 0, 0, 1, 0, 1],
                [1, 0, 0, 1, 0, 1, 0, 1, 0, 1, 1, 1, 0, 0, 0],
                [0, 1, 0, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0],
                [1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0],
                # ---------------------------------------------
                [1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 1],
                [1, 0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0],
                [1, 0, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 1, 0, 0],
                [0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0],
                [1, 0, 0, 1, 1, 1, 1, 0, 1, 0, 0, 1, 1, 0, 1],
                [0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 1, 0],
                [1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1, 1],
            ]
        )
        expect = [
            [1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 1],
            [0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0],
            [0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0],
            # ---------------------------------------------
            [0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 0],
            [0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1],
            [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            # ---------------------------------------------
            [1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 1, 0, 1, 0],
            [0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 1],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ]
        out = abd.mask_multiple_infections_3_chunks(input, split0=4, split1=12).eval()
        self.assertTrue(np.all(np.equal(expect, out)))


class TestIncorporatePCRPos(unittest.TestCase):
    def test_cases(self):
        """
        Each column in these arrays test a different case:

        Column:
            1. PCR+ after infection should take precedence
            2. PCR+ before infection should take precedence
            3. No PCR+ -> i_raw should be unchanged
            4. PCR+ without i_raw -> should match PCR+ row
            5. No PCR+ or infection -> all zeros
            6. Multiple PCR+ -> all PCR+ are retained.
        """
        pcr_pos = np.array(
            [
                [0, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 1],
                [0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 0, 0],
                [1, 0, 0, 0, 0, 1],
                [0, 0, 0, 0, 0, 0],
            ]
        )
        i_raw = np.array(
            [
                [0, 0, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0],
            ]
        )
        expect = np.array(
            [
                [0, 1, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 1],
                [0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 0, 0],
                [1, 0, 0, 0, 0, 1],
                [0, 0, 0, 0, 0, 0],
            ]
        )
        out = abd.incorporate_pcrpos(i_raw=i_raw, pcrpos=pcr_pos).eval()
        self.assertTrue(np.all(np.equal(expect, out)), f"output:\n{np.array(out)}")


class TestTwoTimeChunks(unittest.TestCase):
    def test_split_create_chunks_correct_length(self):
        """
        - 10 total time gaps
        - Split at 4 should create a chunk of length 4 and a chunk of length 6.
        """
        pcrpos = np.array(
            [
                [0, 0],
                [0, 0],
                [0, 0],
                [0, 0],
                [0, 0],
                [0, 0],
                [0, 0],
                [0, 0],
                [0, 0],
                [0, 0],
            ]
        )

        # split at 4, making one length 4 chunk and one length 6 chunk
        ttc = abd.TwoTimeChunks(split=4, pcrpos=at.as_tensor(pcrpos))

        # Test that the sizes of the PCR pos arrays are as expected.
        self.assertEqual(4, ttc.pcrpos_0.shape.eval()[0])
        self.assertEqual(6, ttc.pcrpos_1.shape.eval()[0])

    def test_pcrpos_chunk1_dont_mask_chunk2(self):
        """
        A PCR+ in the first time chunk shouldn't mask inferred infections in the second
        time chunk.

        Also, multiple PCR+ in one chunk should appear in the output.
        """
        pcrpos = [
            [0, 0],
            [1, 0],
            [0, 0],
            [0, 1],  # Chunk 1
            # -------- Split ---
            [0, 0],  # Chunk 2
            [0, 1],
            [0, 0],
            [0, 0],
            [0, 1],
            [0, 0],
        ]
        i_raw = [
            [0, 0],
            [0, 0],
            [0, 0],
            [0, 0],  # Chunk 1
            # -------- Split ---
            [0, 0],  # Chunk 2
            [0, 0],
            [1, 0],
            [0, 0],
            [0, 1],
            [0, 0],
        ]

        ttc = abd.TwoTimeChunks(split=4, pcrpos=at.as_tensor(pcrpos))
        output = ttc.incorporate_pcrpos(i_raw=at.as_tensor(i_raw)).eval()

        expect = [
            [0, 0],
            [1, 0],
            [0, 0],
            [0, 1],  # Chunk 1
            # -------- Split ---
            [0, 0],  # Chunk 2
            [0, 1],
            [1, 0],
            [0, 0],
            [0, 1],
            [0, 0],
        ]

        self.assertTrue(np.all(np.equal(expect, output)))

    def test_multiple_infections_survive_incorporate_pcrpos(self):
        """
        Test that TwoTimeChunks.incorporate_pcrpos
        """
        pcrpos = [
            [0],
            [0],
            [0],
            [1],
            # ------ Split ---
            [0],
            [1],
            [0],
            [0],
            [1],
            [0],
        ]
        i_raw = [
            [0],
            [0],
            [0],
            [0],
            # ------ Split ---
            [0],
            [0],
            [0],
            [0],
            [0],
            [0],
        ]

        ttc = abd.TwoTimeChunks(split=4, pcrpos=at.as_tensor(pcrpos))

        expect = [
            [0],
            [0],
            [0],
            [1],
            # ------ Split ---
            [0],
            [1],
            [0],
            [0],
            [1],
            [0],
        ]

        with pm.Model(coords=dict(gap=range(10), ind=range(2))):
            output = ttc.incorporate_pcrpos(i_raw=at.as_tensor(i_raw)).eval()

        self.assertTrue(
            np.all(np.equal(expect, output)), f"output:\n{np.array(output)}"
        )


class TestTwoTimeChunksConstrainInfections(unittest.TestCase):
    """
    MultipleTimeChunks is the base class for TestTwoTimeChunks.

    MultipleTimeChunks.constrain_infections combines mask_multiple_infections and
    incorporate_pcrpos.
    """

    def test_constrain_infections(self):
        """
        Testing the behaviour of TwoTimeChunks.constrain_infections.

        Columns:
            1: Two PCR+ separated by 3 gaps - expect both in output.
            2: Two PCR+ separated by 2 gaps - expect first in output.
            3: Three PCR+ separated each by 2 gaps - expect first and third.
        """
        pcrpos = [
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 1],
            # ------ Split ---
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 1],
            [0, 0, 0],
            [0, 0, 0],
            [1, 1, 1],
        ]
        i_raw = [
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            # ------ Split ---
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
        ]

        ttc = abd.TwoTimeChunks(split=4, pcrpos=at.as_tensor(pcrpos))

        expect = [
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 1],
            # ------ Split ---
            [0, 0, 0],
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 0],
            [0, 0, 0],
            [1, 0, 1],
        ]

        with pm.Model(coords=dict(gap=range(10), ind=range(3))):
            output = ttc.constrain_infections(i_raw=at.as_tensor(i_raw)).eval()

        self.assertTrue(
            np.all(np.equal(expect, output)), f"output:\n{np.array(output)}"
        )


class TestInvLogistic(unittest.TestCase):
    """
    Tests for abdpymc.logistic and abdpymc.invlogistic.
    """

    def test_invlogistic_is_inverse_of_logistic(self):
        """
        invlogistic should be the inverse of logistic.
        """
        kwds = dict(a=-0.123, b=1.357, d=7.78)
        self.assertEqual(3, abd.logistic(abd.invlogistic(3, **kwds), **kwds))


class TiterDataPath:
    @classmethod
    def setUpClass(cls):
        cls.directory = Path(
            Path(abd.__file__).parent.parent, "data", "test_data", "cohort_data"
        )


class TestTiterData(TiterDataPath, unittest.TestCase):
    """
    Tests for abdpymc.TiterData.
    """

    def test_coords_inds(self):
        """
        `ind` coord should be 0, ..., n_inds - 1
        """
        data = abd.TiterData.from_disk(self.directory)
        self.assertEqual(list(range(10)), list(data.coords["ind"]))

    def test_coords_gaps(self):
        """
        `gap` coord should be 0, ..., n_gaps - 1
        """
        data = abd.TiterData.from_disk(self.directory)
        self.assertEqual(list(range(26)), list(data.coords["gap"]))


class TestModel(TiterDataPath, unittest.TestCase):
    """
    Tests for abdpymc.model.
    """

    def test_test_data(self):
        """
        Test the test data is as expected.
        """
        data = abd.TiterData.from_disk(self.directory)
        self.assertEqual(10, data.n_inds)
        self.assertEqual(26, data.n_gaps)
        self.assertEqual((10, 26), data.vacs.shape)
        self.assertEqual((10, 26), data.pcrpos.shape)

    def test_indexes_no_splits(self):
        """
        Test that the only indexes present are gap and ind.
        """
        data = abd.TiterData.from_disk(self.directory)
        with abd.model(data=data):
            idata = pm.sample_prior_predictive(draws=1)
        self.assertEqual({"chain", "draw", "gap", "ind"}, set(idata.prior.indexes))

    def test_indexes_split_delta(self):
        """
        Test that the only indexes present are gap and ind.
        """
        data = abd.TiterData.from_disk(self.directory)
        splits = data.calculate_splits(delta=True, omicron=False)
        with abd.model(data=data, splits=splits):
            idata = pm.sample_prior_predictive(draws=1)
        self.assertEqual({"chain", "draw", "gap", "ind"}, set(idata.prior.indexes))

    def test_indexes_split_omicron(self):
        """
        Test that the only indexes present are gap and ind.
        """
        data = abd.TiterData.from_disk(self.directory)
        splits = data.calculate_splits(delta=False, omicron=True)
        with abd.model(data=data, splits=splits):
            idata = pm.sample_prior_predictive(draws=1)
        self.assertEqual({"chain", "draw", "gap", "ind"}, set(idata.prior.indexes))

    def test_indexes_split_delta_omicron(self):
        """
        Test that the only indexes present are gap and ind.
        """
        data = abd.TiterData.from_disk(self.directory)
        splits = data.calculate_splits(delta=True, omicron=True)
        with abd.model(data=data, splits=splits):
            idata = pm.sample_prior_predictive(draws=1)
        self.assertEqual({"chain", "draw", "gap", "ind"}, set(idata.prior.indexes))


class TestComputeChunkedCumProbability(unittest.TestCase):
    def test_cum_p_never_exceeds_one(self):
        """
        Cummulative probability should not exceed 1.
        """
        p = np.random.uniform(low=0.5, high=0.8, size=20)
        cum_p = abd.compute_chunked_cum_p(p)
        self.assertEqual(1.0, max(cum_p))

    def test_cum_p_never_exceeds_one_two_chunks(self):
        """
        Cummulative probability should not exceed 1.
        """
        p = np.random.uniform(low=0.7, high=0.8, size=20)
        cum_p = abd.compute_chunked_cum_p(p, splits=(10,))
        self.assertEqual(1.0, max(cum_p[:10]), "value in first chunk exceeds 1.0")
        self.assertEqual(1.0, max(cum_p[10:]), "value in second chunk exceeds 1.0")

    def test_cum_p_two_chunks_probabilities_reset(self):
        """
        Probabilities should go back to being less than one at the start of the second
        chunk.
        """
        p = np.random.uniform(low=0.7, high=0.8, size=20)
        cum_p = abd.compute_chunked_cum_p(p, splits=(10,))
        self.assertLess(cum_p[10], 1.0)

    def test_raises_error_if_p_not_1d(self):
        """
        Should raise a ValueError if p is not 1D.
        """
        with self.assertRaisesRegex(ValueError, "p is not 1D"):
            p = np.random.uniform(0, 1, size=(10, 10))
            abd.compute_chunked_cum_p(p)


class TestTempResponse(unittest.TestCase):
    """
    Tests for abd.abd._temp_response_scalar_rho
    """

    def test_temp_response_shape(self):
        """
        The output shape should match the input shape.
        """
        n_gaps = 10
        n_inds = 5
        exposure = at.as_tensor_variable(np.random.randint(0, 2, size=(n_gaps, n_inds)))
        response = abd.abd._temp_response_scalar_rho(
            exposure, n_inds=n_inds, temp=1.0, rho=0.9
        ).eval()
        self.assertEqual(tuple(exposure.shape.eval()), (n_gaps, n_inds))
        self.assertEqual(response.shape, (n_gaps, n_inds))

    def test_temp_response_values(self):
        """
        Test the correct values are returned.
        """
        exposure = at.as_tensor_variable(
            [
                [1, 0, 0],
                [0, 1, 0],
                [0, 0, 1],
                [1, 0, 0],
                [0, 1, 0],
            ]
        )
        response = abd.abd._temp_response_scalar_rho(
            exposure, n_inds=3, temp=1.0, rho=0.5
        ).eval()
        expected_response = np.array(
            [
                [1.0, 0.0, 0.0],
                [0.5, 1.0, 0.0],
                [0.25, 0.5, 1.0],
                [1.125, 0.25, 0.5],
                [0.5625, 1.125, 0.25],
            ]
        )
        np.testing.assert_almost_equal(response, expected_response)

    def test_temp_response_values_more_gaps_more_individuals(self):
        """
        Like above, but bigger array.
        """
        exposure = at.as_tensor_variable(
            [
                [1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0],
                [0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1],
                [0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0],
                [1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0],
                [0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1],
                [0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0],
                [1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0],
                [0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1],
                [0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0],
                [1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0],
                [0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1],
                [0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0],
                [1, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0],
                [0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 1],
                [0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0],
            ]
        )
        response = abd.abd._temp_response_scalar_rho(
            exposure, n_inds=11, temp=1.0, rho=0.5
        ).eval()
        expected_response = np.array(
            [
                [1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0],
                [0.5, 1.0, 0.0, 0.5, 1.0, 0.0, 0.5, 1.0, 0.0, 0.5, 1.0],
                [0.25, 0.5, 1.0, 0.25, 0.5, 1.0, 0.25, 0.5, 1.0, 0.25, 0.5],
                [1.125, 0.25, 0.5, 1.125, 0.25, 0.5, 1.125, 0.25, 0.5, 1.125, 0.25],
                [
                    0.5625,
                    1.125,
                    0.25,
                    0.5625,
                    1.125,
                    0.25,
                    0.5625,
                    1.125,
                    0.25,
                    0.5625,
                    1.125,
                ],
                [
                    0.28125,
                    0.5625,
                    1.125,
                    0.28125,
                    0.5625,
                    1.125,
                    0.28125,
                    0.5625,
                    1.125,
                    0.28125,
                    0.5625,
                ],
                [
                    1.140625,
                    0.28125,
                    0.5625,
                    1.140625,
                    0.28125,
                    0.5625,
                    1.140625,
                    0.28125,
                    0.5625,
                    1.140625,
                    0.28125,
                ],
                [
                    0.5703125,
                    1.140625,
                    0.28125,
                    0.5703125,
                    1.140625,
                    0.28125,
                    0.5703125,
                    1.140625,
                    0.28125,
                    0.5703125,
                    1.140625,
                ],
                [
                    0.28515625,
                    0.5703125,
                    1.140625,
                    0.28515625,
                    0.5703125,
                    1.140625,
                    0.28515625,
                    0.5703125,
                    1.140625,
                    0.28515625,
                    0.5703125,
                ],
                [
                    1.142578125,
                    0.28515625,
                    0.5703125,
                    1.142578125,
                    0.28515625,
                    0.5703125,
                    1.142578125,
                    0.28515625,
                    0.5703125,
                    1.142578125,
                    0.28515625,
                ],
                [
                    0.5712890625,
                    1.142578125,
                    0.28515625,
                    0.5712890625,
                    1.142578125,
                    0.28515625,
                    0.5712890625,
                    1.142578125,
                    0.28515625,
                    0.5712890625,
                    1.142578125,
                ],
                [
                    0.28564453125,
                    0.5712890625,
                    1.142578125,
                    0.28564453125,
                    0.5712890625,
                    1.142578125,
                    0.28564453125,
                    0.5712890625,
                    1.142578125,
                    0.28564453125,
                    0.5712890625,
                ],
                [
                    1.142822265625,
                    0.28564453125,
                    0.5712890625,
                    1.142822265625,
                    0.28564453125,
                    0.5712890625,
                    1.142822265625,
                    0.28564453125,
                    0.5712890625,
                    1.142822265625,
                    0.28564453125,
                ],
                [
                    0.5714111328125,
                    1.142822265625,
                    0.28564453125,
                    0.5714111328125,
                    1.142822265625,
                    0.28564453125,
                    0.5714111328125,
                    1.142822265625,
                    0.28564453125,
                    0.5714111328125,
                    1.142822265625,
                ],
                [
                    0.28570556640625,
                    0.5714111328125,
                    1.142822265625,
                    0.28570556640625,
                    0.5714111328125,
                    1.142822265625,
                    0.28570556640625,
                    0.5714111328125,
                    1.142822265625,
                    0.28570556640625,
                    0.5714111328125,
                ],
            ]
        )
        np.testing.assert_almost_equal(response, expected_response)

    def test_temp_response_no_exposure(self):
        """
        Test the response when there is no exposure.
        """
        exposure = at.as_tensor_variable(np.zeros((10, 5)))
        response = abd.abd._temp_response_scalar_rho(
            exposure, n_inds=5, temp=1.0, rho=0.9
        ).eval()
        expected_response = np.zeros((10, 5))
        np.testing.assert_almost_equal(response, expected_response)

    def test_compare_to_scan_implementation(self):
        """
        Test that abd.abd._temp_response_scalar_rho gives the same result as abd._temp_response_scan for 5 different
        combinations of random parameters that are passed to these functions.
        """
        np.random.seed(42)
        for _ in range(5):
            n_gaps = np.random.randint(1, 20)
            n_inds = np.random.randint(1, 10)
            exposure = at.as_tensor_variable(
                np.random.randint(0, 2, size=(n_gaps, n_inds))
            )
            temp = np.random.uniform(0.1, 2.0)
            rho = np.random.uniform(0.1, 0.9)

            kwds = dict(exposure=exposure, n_inds=n_inds, temp=temp, rho=rho)

            with self.subTest(
                n_gaps=n_gaps, n_inds=n_inds, temp=temp, rho=rho, exposure=exposure
            ):

                response = abd.abd._temp_response_scalar_rho(**kwds).eval()
                response_scan = abd.abd._temp_response_scan(**kwds).eval()

                np.testing.assert_almost_equal(response, response_scan)

    def test_per_individual_rho(self):
        """
        Test temp_response in a simple case with 3 individuals and 10 gaps where rho is an array
        that contains a different value for each individual.
        """
        exposure = at.as_tensor_variable(
            [
                [1, 0, 0],
                [0, 1, 0],
                [0, 0, 1],
                [1, 0, 0],
                [0, 1, 0],
            ]
        )
        rho = np.array([0.5, 0.6, 0.7])
        response = abd.abd._temp_response_vector_rho(
            exposure, n_inds=3, temp=1.0, rho=rho
        ).eval()
        expected_response = np.array(
            [
                [1.0, 0.0, 0.0],
                [0.5, 1.0, 0.0],
                [0.25, 0.6, 1.0],
                [1.125, 0.36, 0.7],
                [0.5625, 1.216, 0.49],
            ]
        )
        np.testing.assert_almost_equal(response, expected_response)


if __name__ == "__main__":
    # Stop pymc reporting things like which variables are going to be sampled
    logging.getLogger("pymc").setLevel(logging.ERROR)

    unittest.main()



from collections import namedtuple
from numbers import Number
from typing import Optional
import dataclasses

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pydantic import (
    BaseModel,
    ConfigDict,
    field_validator,
    FiniteFloat,
    NegativeFloat,
    NonNegativeFloat,
    PositiveFloat,
)

import abdpymc as abd

InfectionResponses = namedtuple(
    "InfectionResponses", ("infections", "s_response", "n_response")
)


class BaseModelNoExtra(BaseModel):
    """
    Prevent additional fields being passed to a pydantic model.
    """

    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)


class Protection(BaseModelNoExtra):
    """
    Parameters of a protection curve

    Args:
        a: 50% protective titer
        b: Slope of protection curve
    """

    a: FiniteFloat = 0.0
    b: PositiveFloat = 1.0

    def p_protection(self, titer: float) -> float:
        """
        Probability of protection from infection at a given titer.
        """
        return 1 / (1 + np.exp(-self.b * (titer - self.a)))

    def is_protected(self, titer: float) -> float:
        """
        Is an individual with a given titer protected from infection?
        """
        return np.random.uniform() < self.p_protection(titer)

    def plot_curve(self, lo: FiniteFloat = -5, hi: FiniteFloat = 5, **kwds) -> None:
        """
        Plot the protection curve between lo and hi.
        """
        grid = np.linspace(lo, hi)
        plt.plot(grid, self.p_protection(grid), **kwds)


class Elisa(BaseModelNoExtra):
    """
    Parameters of an ELISA titration curve.

    Args:
        b: Slope of the OD sigmoid curve.
        d: Maximum response of the OD curve.
        sd: Standard deviation of the error on an OD reading.
    """

    b: NegativeFloat = -2.2
    d: PositiveFloat = 1.6
    sd: PositiveFloat = 0.1

    def model_od(self, dilution: FiniteFloat, titer: FiniteFloat) -> float:
        """
        What is the measured OD of a sample with a particular titer at a given dilution?

        Args:
            dilution: The dilution of the sample.
            titer: The titer of the sample.
        """
        return np.random.normal(
            loc=abd.logistic(x=dilution, a=titer, b=self.b, d=self.d), scale=self.sd
        )


class Dynamics(BaseModelNoExtra):
    """
    How antibody titers change over time.

    Args:
        init: The initial value.
        perm_rise: Permanent rise after any infection or vaccination.
        temp_rise_{i,v}: Temporary rise after an infection or vaccination.
        temp_wane: Waning rate
    """

    init: FiniteFloat = -2.0
    perm_rise: NonNegativeFloat = 2.0
    temp_rise_i: NonNegativeFloat = 1.5
    temp_rise_v: NonNegativeFloat = 2.0
    temp_wane: PositiveFloat = 0.95

    @field_validator("temp_wane")
    @classmethod
    def wane_must_be_between_0_1(cls, value) -> float:
        if not 0.0 <= value <= 1.0:
            raise ValueError("waning parameter must be between 0-1")
        return value

    def next_temp_response(
        self, prev: float, is_infected: bool, is_vaccinated: bool = False
    ) -> float:
        """
        Calculate the next temporary response given a previous temporary response.

        Args:
            prev: The previous temporary response.
            is_infected: Did this individual just get infected?
            is_vaccinated: Did this individual just get vaccinated?
        """
        return (
            prev * self.temp_wane
            + is_infected * self.temp_rise_i
            + is_vaccinated * self.temp_rise_v
        )

    def perm_response(
        self, infections: np.ndarray, vaccinations: Optional[np.array] = None
    ) -> float:
        """
        Calculate a permanent response, given an array of infections and optional
        vaccinations.

        Args:
            infections: Binary array indicating when infections occurred.
            vaccinations: Binary array indicating when vaccinations occurred. Passing
                None implies that this permanent response is not affected by vaccination
                (i.e. for the N antigen).
        """
        if infections.any():
            return self.perm_rise
        elif vaccinations is not None and vaccinations.any():
            return self.perm_rise
        else:
            return 0.0


class Antibody(BaseModelNoExtra):
    """
    Behaviour of an antibody defined by titer mediated protection from infection, ELISA
    characteristics and how titers change in response to vaccination or infection
    (dynamics).
    """

    protection: Protection = Protection()
    elisa: Elisa = Elisa()
    dynamics: Dynamics = Dynamics()


class Antibodies(BaseModelNoExtra):
    """
    S and N antibodies.
    """

    s: Antibody = Antibody()
    n: Antibody = Antibody()

    def is_protected(self, s_titer: Number, n_titer: Number) -> bool:
        """
        Is an individual protected by either an S or N response?

        Args:
            {s,n}_titer: Titers.
        """
        protected_by_s = self.s.protection.is_protected(s_titer)
        protected_by_n = self.n.protection.is_protected(n_titer)
        return protected_by_n or protected_by_s

    def plot_protection_curves(self, lo=-5, hi=5, **kwds) -> None:
        """
        Plot protection curves for both antibodies.

        Args:
            lo: Titer minimum (x-axis).
            hi: Titer maximum (y-axis).
        """
        self.s.protection.plot_curve(lo=lo, hi=hi, label="S", c=abd.DARKORANGE, **kwds)
        self.n.protection.plot_curve(lo=lo, hi=hi, label="N", c=abd.BLUEGREY, **kwds)


@dataclasses.dataclass
class Individual:
    """
    An individual.

    Args:
        {pcrpos,vacs}: (n_gaps,) binary arrays indicating when individuals had a PCR+
            result or were vaccinated.
        ab: Antibody responses.
    """

    pcrpos: np.ndarray
    vacs: np.ndarray
    ab: Antibodies = dataclasses.field(default_factory=Antibodies)

    def __post_init__(self) -> None:
        self.pcrpos = np.array(self.pcrpos)
        self.vacs = np.array(self.vacs)
        if self.pcrpos.shape != self.vacs.shape:
            raise ValueError("vaccination and pcrpos are different shapes")
        if self.pcrpos.ndim != 1:
            raise ValueError("vaccination and pcrpos should be 1 dimensional")
        self.n_gaps = len(self.pcrpos)

    def infection_responses(self, lam0: np.ndarray) -> InfectionResponses:
        """
        Simulate infections and responses.

        Args:
            lam0: Per-time gap infection probability (baseline infection rate).
        """
        if lam0.ndim != 1:
            raise ValueError("lam0 should be 1D")

        if len(lam0) != self.n_gaps:
            raise ValueError("must have single infection rate for each time gap")

        s = np.empty(self.n_gaps)
        n = np.empty(self.n_gaps)
        infections = np.zeros(self.n_gaps)

        s_temp = 0
        n_temp = 0

        for t in range(self.n_gaps):
            exposed = np.random.uniform() < lam0[t]

            s_prev = s[t - 1] if t > 0 else self.ab.s.dynamics.init
            n_prev = n[t - 1] if t > 0 else self.ab.n.dynamics.init

            if self.pcrpos[t] == 1:
                is_infected = True

            elif exposed and not self.ab.is_protected(s_titer=s_prev, n_titer=n_prev):
                is_infected = True

            else:
                is_infected = False

            if is_infected:
                infections[t] = 1.0

            is_vaccinated = self.vacs[t] == 1.0

            s_temp = self.ab.s.dynamics.next_temp_response(
                s_temp, is_infected=is_infected, is_vaccinated=is_vaccinated
            )
            n_temp = self.ab.n.dynamics.next_temp_response(
                n_temp, is_infected=is_infected
            )

            s_perm = self.ab.s.dynamics.perm_response(
                infections[: t + 1], vaccinations=self.vacs[: t + 1]
            )
            n_perm = self.ab.n.dynamics.perm_response(
                infections[: t + 1], vaccinations=None
            )

            s[t] = self.ab.s.dynamics.init + s_temp + s_perm
            n[t] = self.ab.n.dynamics.init + n_temp + n_perm

        return InfectionResponses(infections=infections, s_response=s, n_response=n)


@dataclasses.dataclass
class Cohort:
    """
    Args:
        random_seed: Passed to np.random.seed.
        cohort_data_path: See abdpymc.abd.TiterData.
        antibodies: Defines antibody responses.

    Attributes:
        n_inds: Number of individuals.
        n_gaps: Number of time gaps.
    """

    random_seed: int
    cohort_data_path: str
    antibodies: Antibodies = dataclasses.field(default_factory=Antibodies)

    def __post_init__(self) -> None:
        np.random.seed(self.random_seed)
        self.true = abd.TiterData.from_disk(self.cohort_data_path)
        self.n_inds, self.n_gaps = self.true.vacs.shape
        self.individuals = [
            Individual(
                pcrpos=self.true.pcrpos[i],
                vacs=self.true.vacs[i],
                ab=self.antibodies,
            )
            for i in range(self.n_inds)
        ]

    def simulate_responses(self, lam0: np.ndarray) -> None:
        """
        Simulate responses for all individuals. This method updates self.s_titer,
        self.n_titer and self.infections.

        Args:
            lam0: (n_gaps,) array containing baseline infection rates for each time gap.
        """
        infection_responses = [
            self.individuals[i].infection_responses(lam0=lam0)
            for i in range(self.n_inds)
        ]
        self.s_titer = np.array([value.s_response for value in infection_responses])
        self.n_titer = np.array([value.n_response for value in infection_responses])
        self.infections = np.array([value.infections for value in infection_responses])

    def simulate_row(self, row: pd.Series) -> dict:
        """
        Simulate an OD reading given this simulation's parameters and a row of data from
        the main cohort dataset (abdpymc.abd.TiterData.data). See
        self.simulate_dataset for more details.

        Args:
            row: pd.Series containing "measurement", "individual_i", "elapsed_months",
                "dilution", "record_id", "log_dilution", "sample_i" and "sample".
        """
        measurement = row["measurement"]
        ab = {"10222020-S": "s", "40588-V08B": "n"}[measurement]
        arr = getattr(self, f"{ab}_titer")
        simulated_titer = arr[row["individual_i"], row["elapsed_months"]]
        elisa = getattr(self.antibodies, ab).elisa
        return dict(
            od=elisa.model_od(dilution=row["log_dilution"], titer=simulated_titer),
            measurement=measurement,
            dilution=row["dilution"],
            record_id=row["record_id"],
            elapsed_months=row["elapsed_months"],
            individual_i=row["individual_i"],
            log_dilution=row["log_dilution"],
            sample_i=row["sample_i"],
            sample=row["sample"],
        )

    def simulate_dataset(self, df_true: pd.DataFrame) -> pd.DataFrame:
        """
        Simulate OD readings for an entire dataset.

        df_true provides the timing of when samples were taken ('elapsed_months' column)
        and the individual this sample was taken from ('individual_i') which are used to
        lookup the associated simulated titers. Sample dilutions and simulated titer are
        then used to model OD measurements. Other columns from df_true are also retained.

        Args:
            df_true: Dataset to model the characteristics of the simulation measurements
                on. Columns must contain "measurement", "individual_i", "elapsed_months",
                "dilution", "record_id", "log_dilution", "sample_i" and "sample".
        """
        return pd.DataFrame([self.simulate_row(row) for _, row in df_true.iterrows()])
